@@ -20,6 +20,9 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 /********************************************************************/
 #include "pari.h"
 #include "paripriv.h"
+
+#define DEBUGLEVEL DEBUGLEVEL_mat
+
 /*******************************************************************/
 /*                                                                 */
 /*                   CHARACTERISTIC POLYNOMIAL                     */
@@ -331,9 +334,7 @@ caradj(GEN x, long v, GEN *py)
   { /* n! not invertible in base ring */
     T = charpoly(x, v);
     if (!py) return gerepileupto(av, T);
-    *py = RgM_adj_from_char(x, v, T);
-    gerepileall(av, 2, &T,py);
-    return T;
+    *py = RgM_adj_from_char(x, v, T); return gc_all(av, 2, &T,py);
   }
   av = avma; y = RgM_shallowcopy(x);
   for (i = 1; i <= n; i++) gcoeff(y,i,i) = gadd(gcoeff(y,i,i), t);
@@ -755,6 +756,22 @@ RgXQ_minpoly_fast(GEN x, GEN y)
 
 #undef code
 
+/* return caract(Mod(x,T)) in variable v */
+GEN
+RgXQ_minpoly(GEN x, GEN T, long v)
+{
+  pari_sp av = avma;
+  GEN R = RgXQ_minpoly_fast(x,T);
+  if (R) { setvarn(R, v); return R; }
+  if (!issquarefree(T))
+  {
+    R = RgXQ_minpoly_naive(x, T);
+    setvarn(R,v); return R;
+  }
+  R = RgXQ_charpoly(x, T, v);
+  return gerepileupto(av, RgX_div(R,RgX_gcd(R, RgX_deriv(R))));
+}
+
 static GEN
 easymin(GEN x, long v)
 {
@@ -768,18 +785,13 @@ easymin(GEN x, long v)
   if (tx == t_POLMOD)
   {
     GEN a = gel(x,2), b = gel(x,1);
+    if (degpol(b)==0) return pol_1(v);
     if (typ(a) != t_POL || varn(a) != varn(b))
     {
       if (varncmp(gvar(a), v) <= 0) pari_err_PRIORITY("minpoly", x, "<", v);
       return deg1pol(gen_1, gneg_i(a), v);
     }
-    R = RgXQ_minpoly_fast(a,b);
-    if (R) { setvarn(R, v); return R; }
-    if (!issquarefree(b))
-    {
-      R = RgXQ_minpoly_naive(a, b);
-      setvarn(R,v); return R;
-    }
+    return RgXQ_minpoly(a, b, v);
   }
   R = easychar(x, v); if (!R) return NULL;
   dR = RgX_deriv(R);  if (!lgpol(dR)) return NULL;
@@ -1011,7 +1023,7 @@ QM_charpoly_ZX_slice(GEN A, GEN dM, GEN P, GEN *mod)
     Hp = Flm_charpoly_i(a, p);
     if (dp != 1) Hp = Flx_rescale(Hp, Fl_inv(dp, p), p);
     Hp = gerepileupto(av, Flx_to_ZX(Hp));
-    *mod = utoi(p); return Hp;
+    *mod = utoipos(p); return Hp;
   }
   T = ZV_producttree(P);
   A = ZM_nv_mod_tree(A, P, T);
@@ -1023,9 +1035,7 @@ QM_charpoly_ZX_slice(GEN A, GEN dM, GEN P, GEN *mod)
     if (dp != 1) gel(H,i) = Flx_rescale(gel(H,i), Fl_inv(dp, p), p);
   }
   H = nxV_chinese_center_tree(H, P, T, ZV_chinesetree(P,T));
-  *mod = gmael(T, lg(T)-1, 1);
-  gerepileall(av, 2, &H, mod);
-  return H;
+  *mod = gmael(T, lg(T)-1, 1); return gc_all(av, 2, &H, mod);
 }
 
 GEN
@@ -1137,7 +1147,7 @@ carberkowitz(GEN x, long v)
     }
     for (i = 1; i <= r+1; i++) gel(V,i) = gel(Q,i);
   }
-  V = RgV_to_RgX(vecreverse(V), v); /* not gtopoly: fail if v > gvar(V) */
+  V = RgV_to_RgX_reverse(V, v); /* not gtopoly: fail if v > gvar(V) */
   V = odd(lx)? gcopy(V): RgX_neg(V);
   return fix_pol(av0, V);
 }
@@ -1468,13 +1478,13 @@ gtrace(GEN x)
       if (ser_isexactzero(x)) return gcopy(x);
       y = cgetg_copy(x, &lx); y[1] = x[1];
       for (i=2; i<lx; i++) gel(y,i) = gtrace(gel(x,i));
-      return normalize(y);
+      return normalizeser(y);
 
     case t_POLMOD:
       y = gel(x,1); z = gel(x,2);
       if (typ(z) != t_POL || varn(y) != varn(z)) return gmulsg(degpol(y), z);
       av = avma;
-      return gerepileupto(av, quicktrace(z, polsym(y, degpol(y)-1)));
+      return gerepileupto(av, RgXQ_trace(z, y));
 
     case t_FFELT:
       y=cgetg(3, t_INTMOD);
@@ -1571,17 +1581,14 @@ static GEN
 gaussred(GEN a, long signature)
 {
   GEN r, ak, al;
-  pari_sp av, av1;
+  pari_sp av = avma, av1;
   long n = lg(a), i, j, k, l, sp, sn, t;
 
   if (typ(a) != t_MAT) pari_err_TYPE("gaussred",a);
   if (n == 1) return signature? mkvec2(gen_0, gen_0): cgetg(1, t_MAT);
   if (lgcols(a) != n) pari_err_DIM("gaussred");
   n--;
-
-  av = avma;
-  r = const_vecsmall(n, 1);
-  av1= avma;
+  r = const_vecsmall(n, 1); av1= avma;
   a = RgM_shallowcopy(a);
   t = n; sp = sn = 0;
   while (t)
@@ -1691,7 +1698,7 @@ rot(GEN x, GEN y, GEN s, GEN u) {
 GEN
 jacobi(GEN a, long prec)
 {
-  pari_sp av1;
+  pari_sp av;
   long de, e, e1, e2, i, j, p, q, l = lg(a);
   GEN c, ja, L, r, L2, r2, unr;
 
@@ -1714,7 +1721,7 @@ jacobi(GEN a, long prec)
     gel(r,j) = cgetg(l,t_COL);
     for (i=1; i<l; i++) gcoeff(r,i,j) = utor(i==j? 1: 0, prec);
   }
-  av1 = avma;
+  av = avma;
 
   e2 = -(long)HIGHEXPOBIT; p = q = 1;
   c = cgetg(l,t_MAT);
@@ -1786,7 +1793,7 @@ jacobi(GEN a, long prec)
   c = indexsort(L);
   r2 = vecpermute(r, c); for (i=1; i<l; i++) gel(r,i) = gel(r2,i);
   L2 = vecpermute(L, c); for (i=1; i<l; i++) gel(L,i) = gel(L2,i);
-  set_avma(av1); return ja;
+  set_avma(av); return ja;
 }
 
 /*************************************************************************/
@@ -1898,8 +1905,7 @@ QM_ImZ_all(GEN x, GEN *U, long remove, long hnf)
 {
   pari_sp av = avma;
   x = QM_ImZ_all_i(x, U, remove, hnf, 0);
-  if (U) gerepileall(av, 2, &x, &U); else x = gerepilecopy(av, x);
-  return x;
+  return gc_all(av, U?2:1, &x, &U);
 }
 GEN
 QM_ImZ_hnfall(GEN x, GEN *U, long remove) { return QM_ImZ_all(x, U, remove, 1); }
