@@ -20,6 +20,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_nf
+
 int new_galois_format = 0;
 
 /* v a t_VEC, lg(v) = 13, sanity check for true rnf */
@@ -33,6 +35,9 @@ rawchecknf(GEN v) { return typ(v)==t_VEC && lg(v)==10; }
 /* v a t_VEC, lg(v) = 11, sanity check for true bnf */
 static int
 v11checkbnf(GEN v) { return rawchecknf(bnf_get_nf(v)); }
+/* v a t_VEC, lg(v) = 12, sanity check for true bnf */
+static int
+v13checkgchar(GEN v) { return rawcheckbnf(gchar_get_bnf(v)); }
 /* v a t_VEC, lg(v) = 10, sanity check for true nf */
 static int
 v10checknf(GEN v) { return typ(gel(v,1))==t_POL; }
@@ -152,14 +157,17 @@ checknfelt_mod(GEN nf, GEN x, const char *s)
   return a;
 }
 
-void
-check_ZKmodule(GEN x, const char *s)
+int
+check_ZKmodule_i(GEN M)
 {
-  if (typ(x) != t_VEC || lg(x) < 3) pari_err_TYPE(s,x);
-  if (typ(gel(x,1)) != t_MAT) pari_err_TYPE(s,gel(x,1));
-  if (typ(gel(x,2)) != t_VEC) pari_err_TYPE(s,gel(x,2));
-  if (lg(gel(x,2)) != lgcols(x)) pari_err_DIM(s);
+  return (typ(M) ==t_VEC && lg(M) >= 3
+          && typ(gel(M,1)) == t_MAT
+          && typ(gel(M,2)) == t_VEC
+          && lgcols(M) == lg(gel(M,2)));
 }
+void
+check_ZKmodule(GEN M, const char *s)
+{ if (!check_ZKmodule_i(M)) pari_err_TYPE(s, M); }
 
 static long
 typv6(GEN x)
@@ -207,6 +215,7 @@ get_bnf(GEN x, long *t)
           if (!v11checkbnf(x)) break;
           *t = typ_BNF; return x;
         case 13:
+          if (v13checkgchar(x)) { *t = typ_GCHAR; return gchar_get_bnf(x); }
           if (!v13checkrnf(x)) break;
           *t = typ_RNF; return NULL;
         case 17: *t = typ_ELL; return NULL;
@@ -250,11 +259,13 @@ get_nf(GEN x, long *t)
           if (!rawchecknf(x = bnf_get_nf(x))) break;
           *t = typ_BNF; return x;
         case 13:
+          if (v13checkgchar(x)) { *t = typ_GCHAR; return gchar_get_nf(x); }
           if (!v13checkrnf(x)) break;
           *t = typ_RNF; return NULL;
         case 17: *t = typ_ELL; return NULL;
       }
       break;
+    case t_QFB: *t = typ_QFB; return NULL;
     case t_COL:
       if (get_prid(x)) { *t = typ_MODPR; return NULL; }
       break;
@@ -273,6 +284,7 @@ nftyp(GEN x)
       switch(lg(x))
       {
         case 13:
+          if (v13checkgchar(x)) return typ_GCHAR;
           if (!v13checkrnf(x)) break;
           return typ_RNF;
         case 10:
@@ -475,16 +487,24 @@ ZX_Q_mul(GEN A, GEN z)
   return gerepilecopy(av, B);
 }
 
-/* pol != 0 in Z[x], returns a monic polynomial POL in Z[x] generating the
- * same field: there exist C in Q, L in Z such that POL(x) = C pol(x/L).
+/* T != 0 in Z[x], returns a monic polynomial U in Z[x] generating the
+ * same field: there exist C in Q, L in Z such that U(x) = C T(x/L).
  * Set *L = NULL if L = 1, and to L otherwise. No garbage collecting. */
 GEN
-ZX_to_monic(GEN pol, GEN *L)
+ZX_to_monic(GEN T, GEN *L)
 {
-  long n = lg(pol)-1;
-  GEN lc = gel(pol,n);
-  if (is_pm1(lc)) { *L = gen_1; return signe(lc) > 0? pol: ZX_neg(pol); }
-  return ZX_primitive_to_monic(Q_primpart(pol), L);
+  GEN lc = leading_coeff(T);
+  if (is_pm1(lc)) { *L = gen_1; return signe(lc) > 0? T: ZX_neg(T); }
+  return ZX_primitive_to_monic(Q_primpart(T), L);
+}
+
+GEN
+poltomonic(GEN T, GEN *L)
+{
+  pari_sp av = avma;
+  if (typ(T) != t_POL || !RgX_is_QX(T)) pari_err_TYPE("poltomonic", T);
+  if (degpol(T) < 0) pari_err_ROOTS0("poltomonic");
+  T = ZX_Q_normalize(Q_primpart(T), L); return gc_all(av, L? 2: 1, &T, L);
 }
 
 GEN
@@ -574,7 +594,7 @@ zk_galoisapplymod(GEN nf, GEN z, GEN S, GEN p)
   {
     ulong e = Z_pvalrem(den, p, &den);
     if (e) { pe = powiu(p, e); pe1 = mulii(pe, p); }
-    denpe = Fp_inv(den, pe1);
+    denpe = Zp_inv(den, p, e+1);
   }
   R = FpX_FpC_nfpoleval(nf, FpX_red(z, pe1), FpC_red(S, pe1), pe1);
   if (denpe) R = FpC_Fp_mul(R, denpe, pe1);
@@ -968,7 +988,7 @@ idealramgroupswild(GEN nf, GEN gal, GEN aut, GEN pr)
   {
     GEN Dg;
     g = nf_to_scalar_or_alg(nf, modpr_genFq(modpr));
-    if (!gcmpX(g)) /* p | nf.index */
+    if (!gequalX(g)) /* p | nf.index */
     {
       g = Q_remove_denom(g, &Dg);
       vDg = Z_pval(Dg,p);
@@ -1172,6 +1192,11 @@ nfisisom(GEN a, GEN b)
   b = get_nfpol(b, &nfb);
   if (!nfa) { a = Q_primpart(a); RgX_check_ZX(a, "nfisisom"); }
   if (!nfb) { b = Q_primpart(b); RgX_check_ZX(b, "nfisisom"); }
+  if (ZX_equal(a, b))
+  {
+    y = galoisconj(nfb? nfb: b, NULL); settyp(y, t_VEC);
+    return gerepilecopy(av,y);
+  }
   if (nfa && !nfb) { swap(a,b); nfb = nfa; nfa = NULL; sw = 1; }
   if (!tests_OK(a, nfa, b, nfb, 1)) { set_avma(av); return gen_0; }
 
@@ -1208,21 +1233,80 @@ partmap_reverse(GEN a, GEN b, GEN t, GEN la, GEN lb, long v)
   return gerepilecopy(av, z);
 }
 
+static GEN
+partmap_reverse_frac(GEN a, GEN b, GEN t, GEN la, GEN lb, long v)
+{
+  pari_sp av = avma;
+  long k = 0;
+  GEN N, D, G, L, de;
+  GEN C = ZX_ZXY_resultant_all(a, Q_remove_denom(t,&de), &k, &L);
+  if (k || degpol(b) != degpol(C))
+    { setvarn(b,v); pari_err_IRREDPOL("nfisincl", b); }
+  L =  Q_primpart(L);
+  N = gel(L,1); if (!signe(N)) { set_avma(av); return pol_0(v); }
+  D = gel(L,2);
+  N = RgX_neg(N); setvarn(N, v); setvarn(D, v);
+  G = QX_gcd(N,D);
+  if (degpol(G)) { N = RgX_div(N,G); D = RgX_div(D,G); }
+  if (!isint1(lb)) { N = RgX_unscale(N, lb); D = RgX_unscale(D, lb); }
+  if (!isint1(la)) D = RgX_Rg_mul(D, la);
+  return gerepilecopy(av, mkrfrac(N,D));
+}
+
+GEN
+partmap_reverse_frac_worker(GEN t, GEN a, GEN b, GEN la, GEN lb, long v)
+{ return partmap_reverse_frac(a, b, t, la, lb, v); }
+
+static GEN
+nfisincl_from_fact(GEN a, long da, GEN b, GEN la, GEN lb, long vb, GEN y,
+                   long flag)
+{
+  long i, k, l = lg(y), db = degpol(b), d = db / da;
+  GEN x = cgetg(l, t_VEC);
+  for (i= k = 1; i < l; i++)
+  {
+    GEN t = gel(y,i);
+    if (degpol(t) != d) continue;
+    gel(x, k++) = partmap_reverse(a, b, t, la, lb, vb);
+    if (flag) return gel(x,1);
+  }
+  if (k==1) return gen_0;
+  setlg(x, k);
+  gen_sort_inplace(x, (void*)&cmp_RgX, &cmp_nodata, NULL);
+  return x;
+}
+
+static GEN
+nfisincl_from_fact_frac(GEN a, GEN b, GEN la, GEN lb, long vb, GEN y)
+{
+  long i, k, l = lg(y), d = degpol(b) / degpol(a);
+  GEN worker, x = cgetg(l, t_VEC);
+  for (i = k = 1; i < l; i++)
+  {
+    GEN t = gel(y,i);
+    if (degpol(t) != d) continue;
+    gel(x, k++) = t;
+  }
+  if (k==1) return gen_0;
+  worker = snm_closure(is_entry("_partmap_reverse_frac_worker"),
+                       mkvec5(a,b,la,lb,stoi(vb)));
+  setlg(x, k); return gen_parapply(worker, x);
+}
+
 GEN
 nfisincl0(GEN fa, GEN fb, long flag)
 {
   pari_sp av = avma;
-  long i, k, vb, lx;
-  long da, db, d;
+  long vb;
   GEN a, b, nfa, nfb, x, y, la, lb;
   int newvar;
-  if (flag < 0 || flag > 1) pari_err_FLAG("nfisincl");
+  if (flag < 0 || flag > 3) pari_err_FLAG("nfisincl");
+
   a = get_nfpol(fa, &nfa);
   b = get_nfpol(fb, &nfb);
-  da = degpol(a); db = degpol(b);
   if (!nfa) { a = Q_primpart(a); RgX_check_ZX(a, "nsisincl"); }
   if (!nfb) { b = Q_primpart(b); RgX_check_ZX(b, "nsisincl"); }
-  if (ZX_equal(a, b))
+  if (ZX_equal(a, b) && flag<=1)
   {
     if (flag==1)
     {
@@ -1232,28 +1316,18 @@ nfisincl0(GEN fa, GEN fb, long flag)
     x = galoisconj(fb, NULL); settyp(x, t_VEC);
     return gerepilecopy(av,x);
   }
-  if (!tests_OK(a, nfa, b, nfb, 0)) { set_avma(av); return gen_0; }
+  if (flag==0 && !tests_OK(a, nfa, b, nfb, 0)) { set_avma(av); return gen_0; }
 
   if (nfb) lb = gen_1; else nfb = b = ZX_Q_normalize(b,&lb);
   if (nfa) la = gen_1; else nfa = a = ZX_Q_normalize(a,&la);
   vb = varn(b); newvar = (varncmp(varn(a),vb) <= 0);
   if (newvar) { b = leafcopy(b); setvarn(b, fetch_var_higher()); }
   y = lift_shallow(gel(nffactor(nfa,b),1));
-  lx = lg(y);
-  da = degpol(a); db = degpol(b); d = db/da;
-  x = cgetg(lx, t_VEC);
-  for (i=1, k=1; i<lx; i++)
-  {
-    GEN t = gel(y,i);
-    if (degpol(t)!=d) continue;
-    gel(x, k++) = partmap_reverse(a, b, t, la, lb, vb);
-    if (flag==1) break;
-  }
+  if (flag==2)
+    x = nfisincl_from_fact_frac(a, b, la, lb, vb, y);
+  else
+    x = nfisincl_from_fact(nfa, degpol(a), b, la, lb, vb, y, flag);
   if (newvar) (void)delete_var();
-  if (k==1) { set_avma(av); return gen_0; }
-  if (flag==1) return gerepilecopy(av,gel(x,1));
-  setlg(x, k);
-  gen_sort_inplace(x, (void*)&cmp_RgX, &cmp_nodata, NULL);
   return gerepilecopy(av,x);
 }
 
@@ -1263,6 +1337,67 @@ nfisincl(GEN fa, GEN fb) { return nfisincl0(fa, fb, 0); }
 static GEN
 lastel(GEN x) { return gel(x, lg(x)-1); }
 
+static GEN
+RgF_to_Flxq(GEN F, GEN T, ulong p)
+{
+  GEN N, D, iD;
+  if (typ(F)==t_POL) return RgX_to_Flx(F, p);
+  N = RgX_to_Flx(gel(F,1), p); D = RgX_to_Flx(gel(F,2), p);
+  iD = Flxq_invsafe(D, T, p);
+  if (!iD) return NULL;
+  return Flxq_mul(N, iD, T, p);
+}
+
+#define pari_APPLY_abort(EXPR)\
+  { \
+    long i, _l; \
+    GEN _y = cgetg_copy(x, &_l);\
+    for (i=1; i<_l; i++) \
+    { GEN _z = EXPR;\
+      if (!_z) return _z;\
+       gel(_y,i) = _z;\
+    } return _y;\
+  }
+
+static GEN
+RgFV_to_FlxqV(GEN x, GEN T, ulong p)
+{ pari_APPLY_abort(RgF_to_Flxq(gel(x,i), T, p)) }
+
+static GEN
+nfsplitting_auto(GEN g, GEN R)
+{
+  pari_sp av;
+  forprime_t T;
+  long i, d = degpol(g);
+  ulong p;
+  GEN P, K, N, G, q, den = Q_denom(R), Rp, Gp;
+  u_forprime_init(&T, d*maxss(expu(d)-3, 2), ULONG_MAX);
+  av = avma;
+  for(;; set_avma(av))
+  {
+    p = u_forprime_next(&T);
+    if (dvdiu(den,p)) continue;
+    Gp = ZX_to_Flx(g, p);
+    if (!Flx_is_totally_split(Gp, p)) continue;
+    P = Flx_roots(Gp, p);
+    Rp = RgFV_to_FlxqV(R, Gp, p);
+    if (Rp) break;
+    if (DEBUGLEVEL) err_printf("nfsplitting_auto: bad p : %lu\n",p);
+  }
+  if (d == 1) return mkvec3(g, mkcol(gel(Rp,1)), utoi(p));
+  K = Flm_Flc_invimage(FlxV_to_Flm(Rp, d), vecsmall_ei(d, 2), p);
+  if (!K) pari_err_BUG("nfsplitting_auto");
+  N = Flm_transpose(FlxV_Flv_multieval(Rp, P, p));
+  q = perm_inv(vecsmall_indexsort(gel(N,1)));
+  G = cgetg(d+1, t_COL);
+  for (i=1; i<=d; i++)
+  {
+    GEN r = perm_mul(vecsmall_indexsort(gel(N,i)), q);
+    gel(G,i) = FlxV_Flc_mul(Rp, vecpermute(K, r), p);
+  }
+  return mkvec3(g, G, utoi(p));
+}
+
 static GEN
 nfsplitting_composite(GEN P)
 {
@@ -1277,55 +1412,68 @@ nfsplitting_composite(GEN P)
   return Q ? Q: pol_x(varn(P));
 }
 GEN
-nfsplitting(GEN T, GEN D)
+nfsplitting0(GEN T0, GEN D, long flag)
 {
   pari_sp av = avma;
-  long d, v;
-  GEN F, K;
-  T = get_nfpol(T,&K);
+  long d, Ds, v;
+  GEN T, F, K, N = NULL, lT = NULL;
+  if (flag < 0 || flag > 3) pari_err_FLAG("nfsplitting");
+  T = T0 = get_nfpol(T0, &K);
   if (!K)
   {
+    GEN c;
     if (typ(T) != t_POL) pari_err_TYPE("nfsplitting",T);
-    T = Q_primpart(T);
+    T = Q_primitive_part(T, &c);
+    lT = leading_coeff(T); if (isint1(lT)) lT = NULL;
+    if (flag && (c || lT)) pari_err_TYPE("nfsplitting", T0);
     RgX_check_ZX(T,"nfsplitting");
   }
   T = nfsplitting_composite(T);
-  d = degpol(T);
-  if (d<=1) return pol_x(varn(T));
+  if (flag && !ZX_equal(T, T0)) pari_err_IRREDPOL("nfsplitting", T0);
+  d = degpol(T); v = varn(T);
+  if (d <= 1 && !flag) { set_avma(av); return pol_x(v); }
   if (!K) {
-    if (!isint1(leading_coeff(T))) K = T = polredbest(T,0);
+    if (lT) T = polredbest(T,0);
     K = T;
   }
   if (D)
-  {
-    if (typ(D) != t_INT || signe(D) < 1) pari_err_TYPE("nfsplitting",D);
-  }
+  { if (typ(D) != t_INT || signe(D) < 1) pari_err_TYPE("nfsplitting",D); }
+  else if (d <= 7 ||
+           (d <= 11 && pari_is_dir(stack_strcat(pari_datadir, "/galdata"))))
+    D = gel(polgalois(T,DEFAULTPREC), 1);
   else
-  {
-    char *data = stack_strcat(pari_datadir, "/galdata");
-    long dmax = pari_is_dir(data)? 11: 7;
-    D = (d <= dmax)? gel(polgalois(T,DEFAULTPREC), 1): mpfact(d);
-  }
-  d = itos_or_0(D);
-  v = varn(T);
+    D = mpfact(d);
+  Ds = itos_or_0(D);
   T = leafcopy(T); setvarn(T, fetch_var_higher());
   for(F = T;;)
   {
     GEN P = gel(nffactor(K, F), 1), Q = gel(P,lg(P)-1);
-    if (degpol(gel(P,1)) == degpol(Q)) break;
+    if (degpol(gel(P,1)) == degpol(Q))
+    {
+      if (!flag) break;
+      P = liftall_shallow(P);
+      if (flag==1)
+        N = nfisincl_from_fact(K, d, F, gen_1, gen_1, v, P, flag);
+      else
+        N = nfisincl_from_fact_frac(T0, F, gen_1, gen_1, v, P);
+      break;
+    }
     F = rnfequation(K,Q);
-    if (degpol(F) == d) break;
+    if (degpol(F) == Ds && !flag) break;
   }
   if (umodiu(D,degpol(F)))
   {
     char *sD = itostr(D);
     pari_warn(warner,stack_strcat("ignoring incorrect degree bound ",sD));
   }
-  (void)delete_var();
-  setvarn(F,v);
+  setvarn(F, v); (void)delete_var();
+  if (flag) F = flag == 3? nfsplitting_auto(F, N): mkvec2(F, N);
   return gerepilecopy(av, F);
 }
 
+GEN
+nfsplitting(GEN T, GEN D) { return nfsplitting0(T, D, 0); }
+
 /*************************************************************************/
 /**                                                                     **/
 /**                               INITALG                               **/
@@ -1368,7 +1516,7 @@ nf_get_allroots(GEN nf)
 }
 
 /* For internal use. compute trace(x mod pol), sym=polsym(pol,deg(pol)-1) */
-GEN
+static GEN
 quicktrace(GEN x, GEN sym)
 {
   GEN p1 = gen_0;
@@ -1560,6 +1708,15 @@ make_G(nffp_t *F)
   F->G = G;
 }
 
+static long
+prec_fix(long prec)
+{
+#ifndef LONG_IS_64BIT
+  /* make sure that default accuracy is the same on 32/64bit */
+  if (odd(prec)) prec += EXTRAPRECWORD;
+#endif
+  return prec;
+}
 static void
 make_M_G(nffp_t *F, int trunc)
 {
@@ -1574,11 +1731,7 @@ make_M_G(nffp_t *F, int trunc)
     if (er < 0) er = 0;
     F->extraprec = nbits2extraprec(n*er + eBD + log2(n));
   }
-  prec = F->prec + F->extraprec;
-#ifndef LONG_IS_64BIT
-  /* make sure that default accuracy is the same on 32/64bit */
-  if (odd(prec)) prec += EXTRAPRECWORD;
-#endif
+  prec = prec_fix(F->prec + F->extraprec);
   if (!F->ro || gprecision(gel(F->ro,1)) < prec)
     F->ro = get_roots(F->T, F->r1, prec);
 
@@ -1738,15 +1891,14 @@ nfcertify(GEN nf)
 static GEN
 get_red_G(nfmaxord_t *S, GEN *pro)
 {
+  pari_sp av = avma;
   GEN G, u, u0 = NULL;
-  pari_sp av;
-  long i, prec, n = degpol(S->T);
+  long prec, n = degpol(S->T);
   nffp_t F;
 
   prec = nbits2prec(n+32);
   nffp_init(&F, S, prec);
-  av = avma;
-  for (i=1; ; i++)
+  for (;;)
   {
     F.prec = prec; make_M_G(&F, 0); G = F.G;
     if (u0) G = RgM_mul(G, u0);
@@ -1771,7 +1923,7 @@ get_red_G(nfmaxord_t *S, GEN *pro)
 /* Compute an LLL-reduced basis for the integer basis of nf(T).
  * set *pro = roots of x if computed [NULL if not computed] */
 static void
-set_LLL_basis(nfmaxord_t *S, GEN *pro, double DELTA)
+set_LLL_basis(nfmaxord_t *S, GEN *pro, long flag, double DELTA)
 {
   GEN B = S->basis;
   long N = degpol(S->T);
@@ -1781,11 +1933,11 @@ set_LLL_basis(nfmaxord_t *S, GEN *pro, double DELTA)
     if (odd(N - S->r1)) pari_err_IRREDPOL("set_LLL_basis", S->T);
   }
   if (!S->basden) S->basden = get_bas_den(B);
+  *pro = NULL; if (flag & nf_NOLLL) return;
   if (S->r1 == N) {
     pari_sp av = avma;
     GEN u = ZM_lll(make_Tr(S), DELTA, LLL_GRAM|LLL_KEEP_FIRST|LLL_IM);
     B = gerepileupto(av, RgV_RgM_mul(B, u));
-    *pro = NULL;
   }
   else
     B = RgV_RgM_mul(B, get_red_G(S, pro));
@@ -1908,6 +2060,15 @@ nfinit_basic_partial(nfmaxord_t *S, GEN T)
   if (typ(T) == t_POL) { nfmaxord(S, mkvec2(T,utoipos(500000)), 0); }
   else nfinit_basic(S, T);
 }
+static void
+nfinit_basic_flag(nfmaxord_t *S, GEN x, long flag)
+{
+  if (flag & nf_PARTIALFACT)
+    nfinit_basic_partial(S, x);
+  else
+    nfinit_basic(S, x);
+}
+
 /* true nf */
 static GEN
 nf_basden(GEN nf)
@@ -1953,7 +2114,7 @@ nfinit_basic(nfmaxord_t *S, GEN T)
 GEN
 nfinit_complete(nfmaxord_t *S, long flag, long prec)
 {
-  GEN nf, unscale;
+  GEN nf, unscale = S->unscale, rev = NULL;
 
   if (!ZX_is_irred(S->T)) pari_err_IRREDPOL("nfinit",S->T);
   if (!(flag & nf_RED) && !ZX_is_monic(S->T0))
@@ -1961,78 +2122,65 @@ nfinit_complete(nfmaxord_t *S, long flag, long prec)
     pari_warn(warner,"nonmonic polynomial. Result of the form [nf,c]");
     flag |= nf_RED | nf_ORIG;
   }
-  unscale = S->unscale;
   if (!(flag & nf_RED) && !isint1(unscale))
   { /* implies lc(x0) = 1 and L := 1/unscale is integral */
     long d = degpol(S->T0);
     GEN L = ginv(unscale); /* x = L^(-deg(x)) x0(L X) */
-    GEN f= powiu(L, (d*(d-1)) >> 1);
+    GEN f = powiu(L, (d*(d-1)) >> 1);
     S->T = S->T0; /* restore original user-supplied x0, unscale data */
     S->unscale = gen_1;
     S->dT    = gmul(S->dT, sqri(f));
-    S->basis   = RgXV_unscale(S->basis, unscale);
+    S->basis = RgXV_unscale(S->basis, unscale);
     S->index = gmul(S->index, f);
   }
   nfmaxord_complete(S); /* more expensive after set_LLL_basis */
   if (flag & nf_RED)
   {
-    GEN ro, rev;
+    GEN ro;
     /* lie to polred: more efficient to update *after* modreverse, than to
      * unscale in the polred subsystem */
     S->unscale = gen_1;
     rev = nfpolred(S, &ro);
     nf = nfmaxord_to_nf(S, ro, prec);
-    if (flag & nf_ORIG)
-    {
-      if (!rev) rev = pol_x(varn(S->T)); /* no improvement */
-      if (!isint1(unscale)) rev = RgX_Rg_div(rev, unscale);
-      nf = mkvec2(nf, mkpolmod(rev, S->T));
-    }
     S->unscale = unscale; /* restore */
-  } else {
-    GEN ro; set_LLL_basis(S, &ro, 0.99);
+  }
+  else
+  {
+    GEN ro; set_LLL_basis(S, &ro, flag, 0.99);
     nf = nfmaxord_to_nf(S, ro, prec);
   }
+  if (flag & nf_ORIG)
+  {
+    if (!rev)
+    { /* no improvement */
+      long v = varn(S->T);
+      rev = degpol(S->T) == 1? pol_0(v): pol_x(v);
+    }
+    if (!isint1(unscale)) rev = RgX_Rg_div(rev, unscale);
+    nf = mkvec2(nf, mkpolmod(rev, S->T));
+  }
   return nf;
 }
 /* Initialize the number field defined by the polynomial x (in variable v)
- * flag & nf_RED:     try a polred first.
- * flag & nf_ORIG
- *    do a polred and return [nfinit(x), Mod(a,red)], where
+ * flag & nf_RED: try a polred first.
+ * flag & nf_ORIG: return [nfinit(x), Mod(a,red)], where
  *    Mod(a,red) = Mod(v,x) (i.e return the base change). */
-GEN
-nfinitall(GEN x, long flag, long prec)
-{
-  const pari_sp av = avma;
-  nfmaxord_t S;
-  GEN nf;
-
-  if (checkrnf_i(x)) return rnf_build_nfabs(x, prec);
-  nfinit_basic(&S, x);
-  nf = nfinit_complete(&S, flag, prec);
-  return gerepilecopy(av, nf);
-}
-
-GEN
-nfinitred(GEN x, long prec)  { return nfinitall(x, nf_RED, prec); }
-GEN
-nfinitred2(GEN x, long prec) { return nfinitall(x, nf_RED|nf_ORIG, prec); }
-GEN
-nfinit(GEN x, long prec)     { return nfinitall(x, 0, prec); }
-
 GEN
 nfinit0(GEN x, long flag,long prec)
 {
-  switch(flag)
-  {
-    case 0:
-    case 1: return nfinitall(x,0,prec);
-    case 2: case 4: return nfinitall(x,nf_RED,prec);
-    case 3: case 5: return nfinitall(x,nf_RED|nf_ORIG,prec);
-    default: pari_err_FLAG("nfinit");
-  }
-  return NULL; /* LCOV_EXCL_LINE */
+  const pari_sp av = avma;
+  nfmaxord_t S;
+  if (flag < 0 || flag > 7) pari_err_FLAG("nfinit");
+  if (checkrnf_i(x)) return rnf_build_nfabs(x, prec);
+  nfinit_basic(&S, x);
+  return gerepilecopy(av, nfinit_complete(&S, flag, prec));
 }
+GEN
+nfinitred(GEN x, long prec)  { return nfinit0(x, nf_RED, prec); }
+GEN
+nfinitred2(GEN x, long prec) { return nfinit0(x, nf_RED|nf_ORIG, prec); }
+GEN
+nfinit(GEN x, long prec)     { return nfinit0(x, 0, prec); }
 
 /* assume x a bnr/bnf/nf */
 long
@@ -2183,7 +2331,7 @@ get_polchar(CG_data *d, GEN x)
 /* Choose a canonical polynomial in the pair { Pmin_a, Pmin_{-a} }, i.e.
  * { z(X), (-1)^(deg z) z(-Z) } and keeping the smallest wrt cmpii_polred
  * Either leave z alone (return 1) or set z <- (-1)^n z(-X). In place. */
-static int
+int
 ZX_canon_neg(GEN z)
 {
   long i, s;
@@ -2249,7 +2397,7 @@ chk_gen(void *data, GEN x)
 
 static long
 chk_gen_prec(long N, long bit)
-{ return nbits2prec(10 + (long)log2((double)N) + bit); }
+{ return prec_fix(nbits2prec(10 + (long)log2((double)N) + bit)); }
 
 /* v = [P,A] two vectors (of ZX and ZV resp.) of same length; remove duplicate
  * polynomials in P, updating A, in place. Among elements having the same
@@ -2287,7 +2435,7 @@ polred_init(nfmaxord_t *S, nffp_t *F, CG_data *d)
   long e, prec, n = degpol(S->T);
   double log2rho;
   GEN ro;
-  set_LLL_basis(S, &ro, 0.9999);
+  set_LLL_basis(S, &ro, 0, 0.9999);
   /* || polchar ||_oo < 2^e ~ 2 (n * rho)^n, rho = max modulus of root */
   log2rho = ro ? (double)gexpo(ro): fujiwara_bound(S->T);
   e = n * (long)(log2rho + log2((double)n)) + 1;
@@ -2422,10 +2570,8 @@ Polred(GEN x, long flag, GEN fa)
   nfmaxord_t S;
   if (fa)
     nfinit_basic(&S, mkvec2(x,fa));
-  else if (flag & nf_PARTIALFACT)
-    nfinit_basic_partial(&S, x);
   else
-    nfinit_basic(&S, x);
+    nfinit_basic_flag(&S, x, flag);
   return gerepilecopy(av, polred_aux(&S, NULL, flag));
 }
 
@@ -2715,7 +2861,7 @@ polredabs_i(GEN x, nfmaxord_t *S, GEN *u, long flag)
   GEN v, y, a;
   long i, l;
 
-  nfinit_basic_partial(S, x);
+  nfinit_basic_flag(S, x, flag);
   x = S->T0;
   if (degpol(x) == 1)
   {
@@ -2724,19 +2870,6 @@ polredabs_i(GEN x, nfmaxord_t *S, GEN *u, long flag)
     return mkvec2(mkvec( pol_x(vx) ),
                   mkvec( deg1pol_shallow(gen_1, negi(gel(S->T,2)), vx) ));
   }
-  if (!(flag & nf_PARTIALFACT) && S->dK && S->dKP)
-  {
-    GEN vw = primes_certify(S->dK, S->dKP);
-    v = gel(vw,1); l = lg(v);
-    if (l != 1)
-    { /* fix integral basis */
-      GEN w = gel(vw,2);
-      for (i = 1; i < l; i++)
-        w = ZV_union_shallow(w, gel(Z_factor(gel(v,i)),1));
-      nfinit_basic(S, mkvec2(S->T0,w));
-    }
-  }
-
   chk.data = (void*)&d;
   polred_init(S, &F, &d);
   d.bound = embed_T2(F.ro, d.r1);
