@@ -21,6 +21,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_mod
+
 /* assume z[1] was created last */
 #define fix_frac_if_int(z) if (equali1(gel(z,2)))\
   z = gerepileupto((pari_sp)(z+3), gel(z,1));
@@ -104,7 +106,7 @@ mulRq(GEN x, GEN y) {
 static GEN
 addqf(GEN x, GEN y, long prec) { pari_sp av = avma;
   long i = gexpo(x) - gexpo(y);
-  if (i > 0) prec += nbits2extraprec( i );
+  if (i > 0) prec += nbits2extraprec(i);
   return gerepileupto(av, gadd(y, quadtofp(x, prec)));
 }
 static GEN
@@ -294,7 +296,9 @@ gred_rfrac_simple(GEN n, GEN d)
   }
   z = cgetg(3,t_RFRAC);
   gel(z,1) = gmul(n, cn);
-  gel(z,2) = rfrac_denom_mul_scal(d, cd);
+  gel(z,2) = d = rfrac_denom_mul_scal(d, cd);
+  /* can happen: Pol(O(17^-1)) / Pol([Mod(9,23), O(23^-3)]) */
+  if (!signe(d)) pari_err_INV("gred_rfrac_simple", d);
   return z;
 }
 
@@ -744,34 +748,35 @@ addsub_polmod_scal(GEN Y, GEN y, GEN x, GEN(*op)(GEN,GEN))
 static GEN
 add_ser_scal(GEN y, GEN x)
 {
-  long i, l, ly, vy;
+  long i, v, ly, vy;
   GEN z;
 
   if (isrationalzero(x)) return gcopy(y);
   ly = lg(y);
-  l = valp(y);
-  if (l < 3-ly) return gcopy(y);
-  /* l + ly >= 3 */
-  if (l < 0)
+  v = valp(y);
+  if (v < 3-ly) return gcopy(y);
+  /* v + ly >= 3 */
+  if (v < 0)
   {
     z = cgetg(ly,t_SER); z[1] = y[1];
-    for (i = 2; i <= 1-l; i++) gel(z,i) = gcopy(gel(y,i));
+    for (i = 2; i <= 1-v; i++) gel(z,i) = gcopy(gel(y,i));
     gel(z,i) = gadd(x,gel(y,i)); i++;
     for (     ; i < ly; i++)   gel(z,i) = gcopy(gel(y,i));
-    return z;
+    return normalizeser(z);
   }
   vy = varn(y);
-  if (l > 0)
+  if (v > 0)
   {
     if (ser_isexactzero(y))
-      return scalarser(ly == 2? x: gadd(x,gel(y,2)), vy, l);
-    y -= l; ly += l;
+      return scalarser(ly == 2? x: gadd(x,gel(y,2)), vy, v);
+    y -= v; ly += v;
     z = cgetg(ly,t_SER);
     x = gcopy(x);
-    for (i=3; i<=l+1; i++) gel(z,i) = gen_0;
+    for (i=3; i<=v+1; i++) gel(z,i) = gen_0;
   }
+  else if (ser_isexactzero(y)) return gcopy(y);
   else
-  { /* l = 0, ly >= 3. Also OK if ser_isexactzero(y) */
+  { /* v = 0, ly >= 3 */
     z = cgetg(ly,t_SER);
     x = gadd(x, gel(y,2));
     i = 3;
@@ -779,7 +784,7 @@ add_ser_scal(GEN y, GEN x)
   for (; i<ly; i++) gel(z,i) = gcopy(gel(y,i));
   gel(z,2) = x;
   z[1] = evalsigne(1) | _evalvalp(0) | evalvarn(vy);
-  return gequal0(x)? normalize(z): z;
+  return gequal0(x)? normalizeser(z): z;
 }
 static long
 _serprec(GEN x) { return ser_isexactzero(x)? 2: lg(x); }
@@ -793,7 +798,10 @@ ser_add(GEN x, GEN y)
   /* valp(x) <= valp(y) */
   lx = _serprec(x);
   if (lx == 2) /* don't lose type information */
-    return scalarser(gadd(Rg_get_0(x), Rg_get_0(y)), varn(x), valp(x));
+  {
+    z = scalarser(gadd(Rg_get_0(x), Rg_get_0(y)), varn(x), 1);
+    setvalp(z, valp(x)); return z;
+  }
   ly = _serprec(y) + n; if (lx < ly) ly = lx;
   if (n)
   {
@@ -805,7 +813,7 @@ ser_add(GEN x, GEN y)
     z = cgetg(ly,t_SER);
     for (i=2; i < ly; i++) gel(z,i) = gadd(gel(x,i),gel(y,i));
   }
-  z[1] = x[1]; return normalize(z);
+  z[1] = x[1]; return normalizeser(z);
 }
 /* typ(y) == RFRAC, x polynomial in same variable or "scalar" */
 static GEN
@@ -922,7 +930,11 @@ add_rfrac(GEN x, GEN y)
   if (!signe(n))
   {
     n = simplify_shallow(n);
-    if (isrationalzero(n)) return gerepileupto(av, zeropol(varn(x2)));
+    if (isexactzero(n))
+    {
+      if (isrationalzero(n)) { set_avma(av); return zeropol(varn(x2)); }
+      return gerepileupto(av, scalarpol(n, varn(x2)));
+    }
     return gerepilecopy(av, mkrfrac(n, RgX_mul(gel(x,2),y2)));
   }
   if (degpol(n) == 0)
@@ -1172,24 +1184,15 @@ gadd(GEN x, GEN y)
     case t_SER:
       if (ty == t_RFRAC)
       {
-        GEN n, d;
         long vn, vd;
         av = avma;
-        n = gel(y,1); vn = gval(n, vy);
-        d = gel(y,2); vd = RgX_valrem(d, &d);
+        vn = gval(gel(y,1), vy);
+        vd = RgX_valrem_inexact(gel(y,2), NULL);
+        if (vd == LONG_MAX) pari_err_INV("gadd", gel(y,2));
 
         l = lg(x) + valp(x) - (vn - vd);
         if (l < 3) { set_avma(av); return gcopy(x); }
-
-        /* take advantage of y = t^n ! */
-        if (degpol(d))
-          y = gdiv(n, RgX_to_ser_inexact(d,l));
-        else {
-          y = gdiv(n, gel(d,2));
-          if (gvar(y) == vy) y = RgX_to_ser(y,l); else y = scalarser(y, vy, l);
-        }
-        setvalp(y, valp(y) - vd);
-        return gerepileupto(av, gadd(y, x));
+        return gerepileupto(av, gadd(x, rfrac_to_ser_i(y, l)));
       }
       break;
   }
@@ -1331,16 +1334,19 @@ gsub(GEN x, GEN y)
 /**                                                                **/
 /********************************************************************/
 static GEN
-mul_ser_scal(GEN y, GEN x) {
+mul_ser_scal(GEN y, GEN x)
+{
   long l, i;
   GEN z;
-  if (isrationalzero(x)) return gmul(Rg_get_0(y), x);
+  if (isexactzero(x)) return gmul(Rg_get_0(y), x);
   if (ser_isexactzero(y))
-    return scalarser(lg(y) == 2? Rg_get_0(x)
-                               : gmul(gel(y,2), x), varn(y), valp(y));
+  {
+    z = scalarser(lg(y) == 2? Rg_get_0(x): gmul(gel(y,2), x), varn(y), 1);
+    setvalp(z, valp(y)); return z;
+  }
   z = cgetg_copy(y, &l); z[1] = y[1];
   for (i = 2; i < l; i++) gel(z,i) = gmul(gel(y,i), x);
-  return normalize(z);
+  return normalizeser(z);
 }
 /* (n/d) * x, x "scalar" or polynomial in the same variable as d
  * [n/d a valid RFRAC]  */
@@ -1387,7 +1393,7 @@ mul_scal(GEN y, GEN x, long ty)
       return RgX_Rg_mul(y, x);
     case t_SER: return mul_ser_scal(y, x);
     case t_RFRAC: return mul_rfrac_scal(gel(y,1),gel(y,2), x);
-    case t_QFI: case t_QFR:
+    case t_QFB:
       if (typ(x) == t_INT && gequal1(x)) return gcopy(y); /* fall through */
   }
   pari_err_TYPE2("*",x,y);
@@ -1787,6 +1793,13 @@ mulcxpowIs(GEN x, long k)
   return x;
 }
 
+static GEN
+init_ser(long l, long v, long e)
+{
+  GEN z = cgetg(l, t_SER);
+  z[1] = evalvalp(e) | evalvarn(v) | evalsigne(1); return z;
+}
+
 /* fill in coefficients of t_SER z from coeffs of t_POL y */
 static GEN
 fill_ser(GEN z, GEN y)
@@ -1798,7 +1811,7 @@ fill_ser(GEN z, GEN y)
     for (i = 2; i < ly; i++) gel(z,i) = gel(y,i);
     for (     ; i < lx; i++) gel(z,i) = gen_0;
   }
-  return normalize(z);
+  return normalizeser(z);
 }
 
 GEN
@@ -1806,7 +1819,7 @@ gmul(GEN x, GEN y)
 {
   long tx, ty, lx, ly, vx, vy, i, l;
   pari_sp av, tetpil;
-  GEN z, p1, p2;
+  GEN z, p1;
 
   if (x == y) return gsqr(x);
   tx = typ(x); ty = typ(y);
@@ -1863,15 +1876,14 @@ gmul(GEN x, GEN y)
       }
       lx = minss(lg(x), lg(y));
       if (lx == 2) return zeroser(vx, valp(x)+valp(y));
-      av = avma; z = cgetg(lx,t_SER);
-      z[1] = evalvalp(valp(x)+valp(y)) | evalvarn(vx) | evalsigne(1);
+      av = avma; z = init_ser(lx, vx, valp(x)+valp(y));
       x = ser2pol_i(x, lx);
       y = ser2pol_i(y, lx);
       y = RgXn_mul(x, y, lx-2);
       return gerepilecopy(av, fill_ser(z,y));
     }
-    case t_QFI: return qficomp(x,y);
-    case t_QFR: return qfrcomp(x,y);
+    case t_VEC: /* handle extended t_QFB */
+    case t_QFB: return qfbcomp(x,y);
     case t_RFRAC: return mul_rfrac(gel(x,1),gel(x,2), gel(y,1),gel(y,2));
     case t_MAT: return RgM_mul(x, y);
 
@@ -2050,23 +2062,20 @@ gmul(GEN x, GEN y)
       {
         case t_SER:
         {
-          long vn;
-          if (lg(x) == 2) return zeropol(vx);
-          if (lg(y) == 2) return zeroser(vx, valp(y)+RgX_val(x));
-          av = avma;
-          vn = RgX_valrem(x, &x);
-          /* take advantage of x = t^n ! */
-          if (degpol(x)) {
-            p1 = RgX_to_ser(x,lg(y));
-            if (vn) settyp(x, t_VECSMALL); /* *new* x left on stack */
-            p2 = gmul(p1,y);
-            settyp(p1, t_VECSMALL); /* p1 left on stack */
-          } else {
-            set_avma(av);
-            p2 = mul_ser_scal(y, gel(x,2));
+          long v;
+          av = avma; v = RgX_valrem(x, &x);
+          if (v == LONG_MAX) return gerepileupto(av, Rg_get_0(x));
+          v += valp(y); ly = lg(y);
+          if (ly == 2) { set_avma(av); return zeroser(vx, v); }
+          if (degpol(x))
+          {
+            z = init_ser(ly, vx, v);
+            x = RgXn_mul(x, ser2pol_i(y, ly), ly-2);
+            return gerepilecopy(av, fill_ser(z, x));
           }
-          setvalp(p2, valp(p2) + vn);
-          return p2;
+          /* take advantage of x = c*t^v */
+          set_avma(av); y = mul_ser_scal(y, gel(x,2));
+          setvalp(y, v); return y;
         }
 
         case t_RFRAC: return mul_rfrac_scal(gel(y,1),gel(y,2), x);
@@ -2197,13 +2206,12 @@ gsqr(GEN x)
         return z;
       }
       if (lx < 40)
-        return normalize( sqr_ser_part(x, 0, lx-3) );
+        return normalizeser( sqr_ser_part(x, 0, lx-3) );
       else
       {
         pari_sp av = avma;
-        GEN z = cgetg(lx,t_SER);
-        z[1] = evalvalp(2*valp(x)) | evalvarn(varn(x)) | evalsigne(1);
-        x = ser2pol_i(x,lx);
+        GEN z = init_ser(lx, varn(x), 2*valp(x));
+        x = ser2pol_i(x, lx);
         x = RgXn_sqr(x, lx-2);
         return gerepilecopy(av, fill_ser(z,x));
       }
@@ -2213,8 +2221,8 @@ gsqr(GEN x)
       gel(z,2) = gsqr(gel(x,2)); return z;
 
     case t_MAT: return RgM_sqr(x);
-    case t_QFR: return qfrsqr(x);
-    case t_QFI: return qfisqr(x);
+    case t_VEC: /* handle extended t_QFB */
+    case t_QFB: return qfbsqr(x);
     case t_VECSMALL:
       z = cgetg_copy(x, &lx);
       for (i=1; i<lx; i++)
@@ -2259,15 +2267,18 @@ div_rfrac(GEN x, GEN y)
 
 /* x != 0 */
 static GEN
-div_ser_scal(GEN y, GEN x) {
+div_ser_scal(GEN y, GEN x)
+{
   long i, l;
   GEN z;
   if (ser_isexactzero(y))
-    return scalarser(lg(y) == 2? Rg_get_0(x)
-                               : gdiv(gel(y,2), x), varn(y), valp(y));
+  {
+    z = scalarser(lg(y) == 2? Rg_get_0(x): gdiv(gel(y,2), x), varn(y), 1);
+    setvalp(z, valp(y)); return z;
+  }
   z = cgetg_copy(y, &l); z[1] = y[1];
   for (i = 2; i < l; i++) gel(z,i) = gdiv(gel(y,i), x);
-  return normalize(z);
+  return normalizeser(z);
 }
 GEN
 ser_normalize(GEN x)
@@ -2318,52 +2329,38 @@ div_scal_T(GEN x, GEN y, long ty) {
   return NULL; /* LCOV_EXCL_LINE */
 }
 
-static GEN
-ser2pol_ii(GEN x, long lx, long y1)
-{
-  GEN y = cgetg(lx, t_POL);
-  long i;
-  for (i = lx-1; i > 1; i--) gel(y,i) = gel(x,i);
-  y[1] = y1; return y;
-}
-
 /* assume tx = ty = t_SER, same variable vx */
 static GEN
 div_ser(GEN x, GEN y, long vx)
 {
-  long v = valp(x) - valp(y), lx = lg(x), ly = lg(y);
-  GEN y_lead, z;
+  long e, v = valp(x) - valp(y), lx = lg(x), ly = lg(y);
+  GEN y0 = y, z;
   pari_sp av = avma;
 
   if (!signe(y)) pari_err_INV("div_ser", y);
   if (ser_isexactzero(x))
   {
     if (lx == 2) return zeroser(vx, v);
-    return scalarser(gmul(gel(x,2),Rg_get_0(y)), varn(x), v);
+    z = scalarser(gmul(gel(x,2),Rg_get_0(y)), varn(x), 1);
+    setvalp(z, v); return z;
   }
-  y_lead = gel(y,2);
-  if (gequal0(y_lead)) /* normalize denominator if leading term is 0 */
+  if (lx < ly) ly = lx;
+  y = ser2pol_i_normalize(y, ly, &e);
+  if (e)
   {
-    GEN y0 = y;
     pari_warn(warner,"normalizing a series with 0 leading term");
-    for (v--, ly--,y++; ly > 2; v--, ly--, y++)
-    {
-      y_lead = gel(y,2);
-      if (!gequal0(y_lead)) break;
-    }
-    if (ly <= 2) pari_err_INV("div_ser", y0);
+    ly -= e; v -= e; if (ly <= 2) pari_err_INV("div_ser", y0);
   }
-  if (ly < lx) lx = ly;
-  z = cgetg(lx,t_SER); z[1] = evalvalp(v) | evalvarn(vx) | evalsigne(1);
-  x = ser2pol_i(x, lx);
-  y = ser2pol_ii(y, lx, z[1] & ~VALPBITS);
-  if (lx == 3) gel(z,2) = gdiv(gel(x,2), gel(y,2));
-  else
-  { /* FIXME: use Karp/Markstein */
-    y = RgXn_mul(x, RgXn_inv_i(y, lx-2), lx-2);
-    z = fill_ser(z,y);
+  z = init_ser(ly, vx, v);
+  if (ly == 3)
+  {
+    gel(z,2) = gdiv(gel(x,2), gel(y,2));
+    if (gequal0(gel(z,2))) setsigne(z, 0); /* can happen: Mod(1,2)/Mod(1,3) */
+    return gerepileupto(av, z);
   }
-  return gerepilecopy(av, z);
+  x = ser2pol_i(x, ly);
+  y = RgXn_div_i(x, y, ly-2);
+  return gerepilecopy(av, fill_ser(z,y));
 }
 /* x,y compatible PADIC */
 static GEN
@@ -2515,8 +2512,8 @@ gdiv(GEN x, GEN y)
       }
       return div_rfrac(x,y);
 
-    case t_QFI: av = avma; return gerepileupto(av, qficomp(x, ginv(y)));
-    case t_QFR: av = avma; return gerepileupto(av, qfrcomp(x, ginv(y)));
+    case t_VEC: /* handle extended t_QFB */
+    case t_QFB: av = avma; return gerepileupto(av, qfbcomp(x, ginv(y)));
 
     case t_MAT:
       av = avma; p1 = RgM_inv(y);
@@ -2781,12 +2778,21 @@ gdiv(GEN x, GEN y)
       switch(ty)
       {
         case t_SER:
-          if (lg(y) == 2)
-            return zeroser(vx, RgX_val(x) - valp(y));
-          p1 = RgX_to_ser(x,lg(y));
-          p2 = div_ser(p1, y, vx);
-          settyp(p1, t_VECSMALL); /* p1 left on stack */
-          return p2;
+        {
+          GEN y0 = y;
+          long v;
+          av = avma; v = RgX_valrem(x, &x);
+          if (v == LONG_MAX) return gerepileupto(av, Rg_get_0(x));
+          v -= valp(y); ly = lg(y); /* > 2 */
+          y = ser2pol_i_normalize(y, ly, &i);
+          if (i)
+          {
+            pari_warn(warner,"normalizing a series with 0 leading term");
+            ly -= i; v -= i; if (ly <= 2) pari_err_INV("gdiv", y0);
+          }
+          z = init_ser(ly, vx, v);
+          return gerepilecopy(av, fill_ser(z, RgXn_div(x, y, ly-2)));
+        }
 
         case t_RFRAC:
         {
@@ -2803,12 +2809,16 @@ gdiv(GEN x, GEN y)
       switch(ty)
       {
         case t_POL:
-          if (lg(x) == 2)
-            return zeroser(vx, valp(x) - RgX_val(y));
-          p1 = RgX_to_ser_inexact(y,lg(x));
-          p2 = div_ser(x, p1, vx);
-          settyp(p1, t_VECSMALL); /* p1 left on stack */
-          return p2;
+        {
+          long v = valp(x);
+          lx = lg(x);
+          if (lx == 2) return zeroser(vx, v - RgX_val(y));
+          av = avma;
+          x = ser2pol_i(x, lx); v -= RgX_valrem_inexact(y, &y);
+          z = init_ser(lx, vx, v);
+          if (!signe(x)) setsigne(z,0);
+          return gerepilecopy(av, fill_ser(z, RgXn_div(x, y, lx - 2)));
+        }
         case t_RFRAC:
           av = avma;
           return gerepileupto(av, gdiv(gmul(x,gel(y,2)), gel(y,1)));
@@ -2820,8 +2830,8 @@ gdiv(GEN x, GEN y)
       {
         case t_POL: return div_rfrac_pol(gel(x,1),gel(x,2), y);
         case t_SER:
-          av = avma; z = RgX_to_ser_inexact(gel(x,2), lg(y));
-          return gerepileupto(av, gdiv(gel(x,1), gmul(z,y)));
+          av = avma;
+          return gerepileupto(av, gdiv(gel(x,1), gmul(gel(x,2), y)));
       }
       break;
   }
@@ -2862,7 +2872,7 @@ gmulsg(long s, GEN y)
       }
       else
       {
-        gel(z,2) = divis(gel(y,2), i);
+        gel(z,2) = diviuexact(gel(y,2), (ulong)i);
         gel(z,1) = mulis(gel(y,1), s/i);
         fix_frac_if_int(z);
       }
@@ -2898,7 +2908,7 @@ gmulsg(long s, GEN y)
       if (!s) return Rg_get_0(y);
       z = cgetg_copy(y, &ly); z[1]=y[1];
       for (i=2; i<ly; i++) gel(z,i) = gmulsg(s,gel(y,i));
-      return normalize(z);
+      return normalizeser(z);
 
     case t_RFRAC:
       if (!s) return zeropol(varn(gel(y,2)));
@@ -2915,6 +2925,86 @@ gmulsg(long s, GEN y)
   return NULL; /* LCOV_EXCL_LINE */
 }
 
+GEN
+gmulug(ulong s, GEN y)
+{
+  long ly, i;
+  pari_sp av;
+  GEN z;
+
+  switch(typ(y))
+  {
+    case t_INT:  return mului(s,y);
+    case t_REAL: return s? mulur(s,y): gen_0; /* gmul semantic */
+    case t_INTMOD: { GEN p = gel(y,1);
+      z = cgetg(3,t_INTMOD);
+      gel(z,2) = gerepileuptoint((pari_sp)z, modii(mului(s,gel(y,2)), p));
+      gel(z,1) = icopy(p); return z;
+    }
+    case t_FFELT: return FF_Z_mul(y,utoi(s));
+    case t_FRAC:
+      if (!s) return gen_0;
+      z = cgetg(3,t_FRAC);
+      i = ugcd(s, umodiu(gel(y,2), s));
+      if (i == 1)
+      {
+        gel(z,2) = icopy(gel(y,2));
+        gel(z,1) = muliu(gel(y,1), s);
+      }
+      else
+      {
+        gel(z,2) = diviuexact(gel(y,2), i);
+        gel(z,1) = muliu(gel(y,1), s/i);
+        fix_frac_if_int(z);
+      }
+      return z;
+
+    case t_COMPLEX:
+      if (!s) return gen_0;
+      z = cgetg(3, t_COMPLEX);
+      gel(z,1) = gmulug(s,gel(y,1));
+      gel(z,2) = gmulug(s,gel(y,2)); return z;
+
+    case t_PADIC:
+      if (!s) return gen_0;
+      av = avma; return gerepileupto(av, mulpp(cvtop2(utoi(s),y), y));
+
+    case t_QUAD: z = cgetg(4, t_QUAD);
+      gel(z,1) = ZX_copy(gel(y,1));
+      gel(z,2) = gmulug(s,gel(y,2));
+      gel(z,3) = gmulug(s,gel(y,3)); return z;
+
+    case t_POLMOD:
+      retmkpolmod(gmulug(s,gel(y,2)), RgX_copy(gel(y,1)));
+
+    case t_POL:
+      if (!signe(y)) return RgX_copy(y);
+      if (!s) return scalarpol(Rg_get_0(y), varn(y));
+      z = cgetg_copy(y, &ly); z[1]=y[1];
+      for (i=2; i<ly; i++) gel(z,i) = gmulug(s,gel(y,i));
+      return normalizepol_lg(z, ly);
+
+    case t_SER:
+      if (ser_isexactzero(y)) return gcopy(y);
+      if (!s) return Rg_get_0(y);
+      z = cgetg_copy(y, &ly); z[1]=y[1];
+      for (i=2; i<ly; i++) gel(z,i) = gmulug(s,gel(y,i));
+      return normalizeser(z);
+
+    case t_RFRAC:
+      if (!s) return zeropol(varn(gel(y,2)));
+      if (s == 1) return gcopy(y);
+      return mul_rfrac_scal(gel(y,1), gel(y,2), utoi(s));
+
+    case t_VEC: case t_COL: case t_MAT:
+      z = cgetg_copy(y, &ly);
+      for (i=1; i<ly; i++) gel(z,i) = gmulug(s,gel(y,i));
+      return z;
+  }
+  pari_err_TYPE("gmulsg",y);
+  return NULL; /* LCOV_EXCL_LINE */
+}
+
 /********************************************************************/
 /**                                                                **/
 /**                       SIMPLE DIVISION                          **/
@@ -2998,6 +3088,100 @@ gdivgs(GEN x, long s)
   return NULL; /* LCOV_EXCL_LINE */
 }
 
+GEN
+gdivgu(GEN x, ulong s)
+{
+  long tx = typ(x), lx, i;
+  pari_sp av;
+  GEN z;
+
+  if (!s)
+  {
+    if (is_matvec_t(tx) && lg(x) == 1) return gcopy(x);
+    pari_err_INV("gdivgu",gen_0);
+  }
+  switch(tx)
+  {
+    case t_INT: return Qdiviu(x, s);
+    case t_REAL: return divru(x,s);
+
+    case t_INTMOD:
+      z = cgetg(3, t_INTMOD); s = umodui(s, gel(x,1));
+      return div_intmod_same(z, gel(x,1), gel(x,2), utoi(s));
+
+    case t_FFELT: return FF_Z_Z_muldiv(x,gen_1,utoi(s));
+
+    case t_FRAC: z = cgetg(3, t_FRAC);
+      i = ugcd(s, umodiu(gel(x,1), s));
+      if (i == 1)
+      {
+        gel(z,2) = mului(s, gel(x,2));
+        gel(z,1) = icopy(gel(x,1));
+      }
+      else
+      {
+        gel(z,2) = mului(s/i, gel(x,2));
+        gel(z,1) = divis(gel(x,1), i);
+      }
+      normalize_frac(z);
+      fix_frac_if_int(z); return z;
+
+    case t_COMPLEX: z = cgetg(3, t_COMPLEX);
+      gel(z,1) = gdivgu(gel(x,1),s);
+      gel(z,2) = gdivgu(gel(x,2),s); return z;
+
+    case t_PADIC: /* divpT */
+    {
+      GEN p = gel(x,2);
+      if (!signe(gel(x,4))) return zeropadic(p, valp(x) - u_pval(s,p));
+      av = avma;
+      return gerepileupto(av, divpp(x, cvtop2(utoi(s),x)));
+    }
+
+    case t_QUAD: z = cgetg(4, t_QUAD);
+      gel(z,1) = ZX_copy(gel(x,1));
+      gel(z,2) = gdivgu(gel(x,2),s);
+      gel(z,3) = gdivgu(gel(x,3),s); return z;
+
+    case t_POLMOD:
+      retmkpolmod(gdivgu(gel(x,2),s), RgX_copy(gel(x,1)));
+
+    case t_RFRAC:
+      if (s == 1) return gcopy(x);
+      return div_rfrac_scal(x, utoi(s));
+
+    case t_POL: case t_SER:
+      z = cgetg_copy(x, &lx); z[1] = x[1];
+      for (i=2; i<lx; i++) gel(z,i) = gdivgu(gel(x,i),s);
+      return z;
+    case t_VEC: case t_COL: case t_MAT:
+      z = cgetg_copy(x, &lx);
+      for (i=1; i<lx; i++) gel(z,i) = gdivgu(gel(x,i),s);
+      return z;
+  }
+  pari_err_TYPE2("/",x, utoi(s));
+  return NULL; /* LCOV_EXCL_LINE */
+}
+
+/* x / (i*(i+1)) */
+GEN
+divrunextu(GEN x, ulong i)
+{
+  if (i & HIGHMASK) /* i(i+1) >= 2^BITS_IN_LONG*/
+    return divri(x, muluu(i , i+1));
+  else
+    return divru(x, i*(i+1));
+}
+/* x / (i*(i+1)) */
+GEN
+gdivgunextu(GEN x, ulong i)
+{
+  if (i & HIGHMASK) /* i(i+1) >= 2^BITS_IN_LONG*/
+    return gdivgu(x, i*(i+1));
+  else
+    return gdiv(x, muluu(i, i+1));
+}
+
 /* True shift (exact multiplication by 2^n) */
 GEN
 gmul2n(GEN x, long n)
@@ -3063,7 +3247,7 @@ gmul2n(GEN x, long n)
       if (ser_isexactzero(x)) return gcopy(x);
       z = cgetg_copy(x, &lx); z[1] = x[1];
       for (i=2; i<lx; i++) gel(z,i) = gmul2n(gel(x,i),n);
-      return normalize(z); /* needed if char = 2 */
+      return normalizeser(z); /* needed if char = 2 */
     case t_VEC: case t_COL: case t_MAT:
       z = cgetg_copy(x, &lx);
       for (i=1; i<lx; i++) gel(z,i) = gmul2n(gel(x,i),n);
@@ -3140,7 +3324,7 @@ ginv(GEN x)
       return gerepile(av,tetpil,divcR(p2,p1));
 
     case t_QUAD:
-      av=avma; p1=gnorm(x); p2=conj_i(x); tetpil=avma;
+      av=avma; p1=quadnorm(x); p2=conj_i(x); tetpil=avma;
       return gerepile(av,tetpil,gdiv(p2,p1));
 
     case t_PADIC: z = cgetg(5,t_PADIC);
@@ -3148,7 +3332,7 @@ ginv(GEN x)
       z[1] = _evalprecp(precp(x)) | evalvalp(-valp(x));
       gel(z,2) = icopy(gel(x,2));
       gel(z,3) = icopy(gel(x,3));
-      gel(z,4) = Fp_inv(gel(x,4),gel(z,3)); return z;
+      gel(z,4) = Zp_inv(gel(x,4),gel(z,2),precp(x)); return z;
 
     case t_POLMOD: return inv_polmod(gel(x,1), gel(x,2));
     case t_FFELT: return FF_inv(x);
@@ -3176,19 +3360,9 @@ ginv(GEN x)
       return z;
     }
 
-    case t_QFR:
-      av = avma; z = cgetg(5, t_QFR);
-      gel(z,1) = gel(x,1);
-      gel(z,2) = negi( gel(x,2) );
-      gel(z,3) = gel(x,3);
-      gel(z,4) = negr( gel(x,4) );
-      return gerepileupto(av, redreal(z));
-
-    case t_QFI:
-      y = gcopy(x);
-      if (!equalii(gel(x,1),gel(x,2)) && !equalii(gel(x,1),gel(x,3)))
-        togglesign(gel(y,2));
-      return y;
+    case t_VEC: /* handle extended t_QFB */
+    case t_QFB:
+      return qfbpow(x, gen_m1);
     case t_MAT:
       y = RgM_inv(x);
       if (!y) pari_err_INV("ginv",x);
