@@ -29,39 +29,21 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 GEN
 map_proto_G(GEN (*f)(GEN), GEN x)
 {
-  if (is_matvec_t(typ(x)))
-  {
-    long lx, i;
-    GEN y = cgetg_copy(x, &lx);
-    for (i=1; i<lx; i++) gel(y,i) = map_proto_G(f, gel(x,i));
-    return y;
-  }
+  if (is_matvec_t(typ(x))) pari_APPLY_same(map_proto_G(f, gel(x,i)));
   return f(x);
 }
 
 GEN
 map_proto_lG(long (*f)(GEN), GEN x)
 {
-  if (is_matvec_t(typ(x)))
-  {
-    long lx, i;
-    GEN y = cgetg_copy(x, &lx);
-    for (i=1; i<lx; i++) gel(y,i) = map_proto_lG(f, gel(x,i));
-    return y;
-  }
+  if (is_matvec_t(typ(x))) pari_APPLY_same(map_proto_lG(f, gel(x,i)));
   return stoi(f(x));
 }
 
 GEN
 map_proto_lGL(long (*f)(GEN,long), GEN x, long y)
 {
-  if (is_matvec_t(typ(x)))
-  {
-    long l, i;
-    GEN t = cgetg_copy(x, &l);
-    for (i=1; i<l; i++) gel(t,i) = map_proto_lGL(f,gel(x,i),y);
-    return t;
-  }
+  if (is_matvec_t(typ(x))) pari_APPLY_same(map_proto_lGL(f,gel(x,i),y));
   return stoi(f(x,y));
 }
 
@@ -175,6 +157,35 @@ gtolong(GEN x)
 /*                         COMPARISONS                             */
 /*                                                                 */
 /*******************************************************************/
+static void
+chk_true_err()
+{
+  GEN E = pari_err_last();
+  switch(err_get_num(E))
+  {
+    case e_STACK: case e_MEM: case e_ALARM:
+      pari_err(0, E); /* rethrow */
+  }
+}
+/* x - y == 0 or undefined */
+static int
+gequal_try(GEN x, GEN y)
+{
+  int i;
+  pari_CATCH(CATCH_ALL) { chk_true_err(); return 0; }
+  pari_TRY { i = gequal0(gadd(x, gneg_i(y))); } pari_ENDCATCH;
+  return i;
+}
+/* x + y == 0 or undefined */
+static int
+gmequal_try(GEN x, GEN y)
+{
+  int i;
+  pari_CATCH(CATCH_ALL) { chk_true_err(); return 0; }
+  pari_TRY { i = gequal0(gadd(x, y)); } pari_ENDCATCH;
+  return i;
+}
+
 int
 isexactzero(GEN g)
 {
@@ -325,21 +336,24 @@ gequal0(GEN x)
   return 0;
 }
 
-/* x a t_POL or t_SER, considered as having valuation v; let X(t) = t^(-v) x(t)
- * return 1 (true) if coeff(X,i) = 0 for all i != 0 and test(coeff(X, 0))
- * is true. Return 0 (false) otherwise, or if x == 0 */
+/* x a t_POL or t_SER, return 1 if test(coeff(X,d)) is true and
+ * coeff(X,i) = 0 for all i != d. Return 0 (false) otherwise */
 static int
-is_monomial_test(GEN x, long v, int(*test)(GEN))
+is_monomial_test(GEN x, long d, int(*test)(GEN))
 {
-  long d, i, l;
-  if (v > 0) return 0;
-  l = lg(x);
-  if (!signe(x))
-  {
-    if (typ(x) == t_SER) return v + l - 2 <= 0;
-    return l > 2 && test(gel(x,2)); /* e.g. O(2^-1)*x^0 */
+  long i, l = lg(x);
+  if (typ(x) == t_SER)
+  { /* "0" * x^v * (1+O(x)) ?  v <= 0 or null ring */
+    if (l == 3 && isexactzero(gel(x,2))) return d >= 2 || test(gel(x,2));
+    if (d < 2) return 0; /* v > 0 */
   }
-  d = 2-v; if (l <= d || !test(gel(x,d))) return 0;
+  if (d >= l)
+  {
+    if (typ(x) == t_POL) return 0; /* l = 2 */
+    /* t_SER, v = 2-d <= 0 */
+    if (!signe(x)) return 1;
+  }
+  else if (!test(gel(x,d))) return 0;
   for (i = 2; i < l; i++) /* 2 <= d < l */
     if (i != d && !gequal0(gel(x,i))) return 0;
   return 1;
@@ -405,8 +419,8 @@ gequal1(GEN x)
     case t_QUAD:
       return gequal1(gel(x,2)) && gequal0(gel(x,3));
 
-    case t_POL: return is_monomial_test(x, 0, &gequal1);
-    case t_SER: return is_monomial_test(x, valp(x), &gequal1);
+    case t_POL: return is_monomial_test(x, 2, &gequal1);
+    case t_SER: return is_monomial_test(x, 2 - valp(x), &gequal1);
 
     case t_RFRAC: return gequal(gel(x,1), gel(x,2));
     case t_COL: return col_test(x, &gequal1);
@@ -455,11 +469,11 @@ gequalm1(GEN x)
     case t_POLMOD:
       return !degpol(gel(x,1)) || gequalm1(gel(x,2));
 
-    case t_POL: return is_monomial_test(x, 0, &gequalm1);
-    case t_SER: return is_monomial_test(x, valp(x), &gequalm1);
+    case t_POL: return is_monomial_test(x, 2, &gequalm1);
+    case t_SER: return is_monomial_test(x, 2 - valp(x), &gequalm1);
 
     case t_RFRAC:
-      av = avma; return gc_bool(av, gequal(gel(x,1), gneg_i(gel(x,2))));
+      av = avma; return gc_bool(av, gmequal_try(gel(x,1), gel(x,2)));
     case t_COL: return col_test(x, &gequalm1);
     case t_MAT: return mat_test(x, &gequalm1);
   }
@@ -518,6 +532,13 @@ cmp_universal(GEN x, GEN y)
       return 0;
 
     case t_POL:
+    {
+      long X = x[1] & (VARNBITS|SIGNBITS);
+      long Y = y[1] & (VARNBITS|SIGNBITS);
+      if (X < Y) return -1;
+      if (X > Y) return 1;
+      return cmp_universal_rec(x, y, 2);
+    }
     case t_SER:
     case t_FFELT:
     case t_CLOSURE:
@@ -929,9 +950,7 @@ gidentical(GEN x, GEN y)
     case t_FFELT:
       return FF_equal(x,y);
 
-    case t_QFR:
-          if (!identicalrr(gel(x,4),gel(y,4))) return 0; /* fall through */
-    case t_QFI:
+    case t_QFB:
       return equalii(gel(x,1),gel(y,1))
           && equalii(gel(x,2),gel(y,2))
           && equalii(gel(x,3),gel(y,3));
@@ -963,7 +982,8 @@ static int
 polequal(GEN x, GEN y)
 {
   long lx, ly;
-  if (signe(x) != signe(y)) return 0;
+  /* Can't do that: Mod(0,1)*x^0 == x^0
+  if (signe(x) != signe(y)) return 0; */
   lx = lg(x); ly = lg(y);
   while (lx > ly) if (!gequal0(gel(x,--lx))) return 0;
   while (ly > lx) if (!gequal0(gel(y,--ly))) return 0;
@@ -998,24 +1018,6 @@ vecequal(GEN x, GEN y)
   return 1;
 }
 
-static int
-gequal_try(GEN x, GEN y)
-{
-  int i;
-  pari_CATCH(CATCH_ALL) {
-    GEN E = pari_err_last();
-    switch(err_get_num(E))
-    {
-      case e_STACK: case e_MEM: case e_ALARM:
-        pari_err(0, E); /* rethrow */
-    }
-    return 0;
-  } pari_TRY {
-    i = gequal0(gadd(x, gneg_i(y)));
-  } pari_ENDCATCH;
-  return i;
-}
-
 int
 gequal(GEN x, GEN y)
 {
@@ -1057,8 +1059,7 @@ gequal(GEN x, GEN y)
       case t_FFELT:
         return FF_equal(x,y);
 
-      case t_QFR:
-      case t_QFI:
+      case t_QFB:
         return equalii(gel(x,1),gel(y,1))
             && equalii(gel(x,2),gel(y,2))
             && equalii(gel(x,3),gel(y,3));
@@ -1113,6 +1114,44 @@ cx_approx_equal(GEN a, GEN b)
   d = gsub(a,b);
   return gc_bool(av, gequal0(d) || (typ(d)==t_COMPLEX && gequal0(cxnorm(d))));
 }
+static int
+r_approx0(GEN x, long e) { return e - expo(x) > bit_prec(x); }
+/* x ~ 0 compared to reference y */
+int
+cx_approx0(GEN x, GEN y)
+{
+  GEN a, b;
+  long e;
+  switch(typ(x))
+  {
+    case t_COMPLEX:
+      a = gel(x,1); b = gel(x,2);
+      if (typ(a) != t_REAL)
+      {
+        if (!gequal0(a)) return 0;
+        a = NULL;
+      }
+      else if (!signe(a)) a = NULL;
+      if (typ(b) != t_REAL)
+      {
+        if (!gequal0(b)) return 0;
+        if (!a) return 1;
+        b = NULL;
+      }
+      else if (!signe(b))
+      {
+        if (!a) return 1;
+        b = NULL;
+      }
+      /* a or b is != NULL iff it is non-zero t_REAL; one of them is */
+      e = gexpo(y);
+      return (!a || r_approx0(a, e)) && (!b || r_approx0(b, e));
+    case t_REAL:
+      return !signe(x) || r_approx0(x, gexpo(y));
+    default:
+      return gequal0(x);
+  }
+}
 /*******************************************************************/
 /*                                                                 */
 /*                          VALUATION                              */
@@ -1192,9 +1231,18 @@ Q_lvalrem(GEN x, ulong p, GEN *y)
 long
 gvaluation(GEN x, GEN p)
 {
-  long tx = typ(x), tp = typ(p);
+  long tx = typ(x), tp;
   pari_sp av;
 
+  if (!p)
+    switch(tx)
+    {
+      case t_PADIC: return valp(x);
+      case t_POL: return RgX_val(x);
+      case t_SER: return valp(x);
+      default: pari_err_TYPE("gvaluation", x);
+    }
+  tp  = typ(p);
   switch(tp)
   {
     case t_INT:
@@ -1263,7 +1311,12 @@ gvaluation(GEN x, GEN p)
         if (vp == vx)
         {
           long val;
-          if (RgX_is_monomial(p)) return RgX_val(x) / degpol(p);
+          if (RgX_is_monomial(p))
+          {
+            val = RgX_val(x); if (val == LONG_MAX) return LONG_MAX;
+            return val / degpol(p);
+          }
+          if (!signe(x)) return LONG_MAX;
           av = avma;
           for (val=0; ; val++)
           {
@@ -1960,9 +2013,9 @@ is_negative(GEN x) {
 GEN
 gabs(GEN x, long prec)
 {
-  long lx, i;
-  pari_sp av, tetpil;
-  GEN y,p1;
+  long lx;
+  pari_sp av;
+  GEN y, N;
 
   switch(typ(x))
   {
@@ -1973,21 +2026,20 @@ gabs(GEN x, long prec)
       return absfrac(x);
 
     case t_COMPLEX:
-      av=avma; p1=cxnorm(x);
-      switch(typ(p1))
+      av=avma; N=cxnorm(x);
+      switch(typ(N))
       {
         case t_INT:
-          if (!Z_issquareall(p1, &y)) break;
+          if (!Z_issquareall(N, &y)) break;
           return gerepileupto(av, y);
         case t_FRAC: {
           GEN a,b;
-          if (!Z_issquareall(gel(p1,1), &a)) break;
-          if (!Z_issquareall(gel(p1,2), &b)) break;
+          if (!Z_issquareall(gel(N,1), &a)) break;
+          if (!Z_issquareall(gel(N,2), &b)) break;
           return gerepileupto(av, gdiv(a,b));
         }
       }
-      tetpil=avma;
-      return gerepile(av,tetpil,gsqrt(p1,prec));
+      return gerepileupto(av, gsqrt(N,prec));
 
     case t_QUAD:
       av = avma;
@@ -1995,7 +2047,7 @@ gabs(GEN x, long prec)
 
     case t_POL:
       lx = lg(x); if (lx<=2) return RgX_copy(x);
-      return is_negative(gel(x,lx-1))? gneg(x): RgX_copy(x);
+      return is_negative(gel(x,lx-1))? RgX_neg(x): RgX_copy(x);
 
     case t_SER:
      if (!signe(x)) pari_err_DOMAIN("abs", "argument", "=", gen_0, x);
@@ -2003,9 +2055,7 @@ gabs(GEN x, long prec)
      return is_negative(gel(x,2))? gneg(x): gcopy(x);
 
     case t_VEC: case t_COL: case t_MAT:
-      y = cgetg_copy(x, &lx);
-      for (i=1; i<lx; i++) gel(y,i) = gabs(gel(x,i),prec);
-      return y;
+      pari_APPLY_same(gabs(gel(x,i),prec));
   }
   pari_err_TYPE("gabs",x);
   return NULL; /* LCOV_EXCL_LINE */
@@ -2359,20 +2409,39 @@ gaffect(GEN x, GEN y)
 GEN
 quadtofp(GEN x, long prec)
 {
-  GEN b, mb2, c, z, Q, u = gel(x,2), v = gel(x,3);
+  GEN b, D, z, u = gel(x,2), v = gel(x,3);
   pari_sp av;
   if (prec < LOWDEFAULTPREC) prec = LOWDEFAULTPREC;
   if (isintzero(v)) return cxcompotor(u, prec);
-  av = avma; Q = gel(x,1); c = gel(Q,2); b = gel(Q,3);
-  z = sqrtr_abs(itor(quad_disc(x), prec));
-  shiftr_inplace(z, -1);
-  mb2 = signe(b) ? real2n(-1, prec): NULL;
-  if (signe(c) > 0) /* c = -D/4 or (1-D)/4 */
-    z = mkcomplex(mb2? mb2: real_0(prec), z);
-  else if (mb2)
-    z = addrr(z, mb2);
-  /* z = (-b + sqrt(D)) / 2 */
-  return gerepileupto(av, gadd(u, gmul(v,z)));
+  av = avma; D = quad_disc(x); b = gel(gel(x,1),3); /* 0 or -1 */
+  /* u + v (-b + sqrt(D)) / 2 */
+  if (!signe(b)) b = NULL;
+  if (b) u = gadd(gmul2n(u,1), v);
+  z = sqrtr_abs(itor(D, prec));
+  if (!b) shiftr_inplace(z, -1);
+  z = gmul(v, z);
+  if (signe(D) < 0)
+  {
+    z = mkcomplex(cxcompotor(u, prec), z);
+    if (!b) return gerepilecopy(av, z);
+    z = gmul2n(z, -1);
+  }
+  else
+  { /* if (b) x ~ (u + z) / 2 and quadnorm(x) ~ (u^2 - z^2) / 4
+     * else x ~ u + z and quadnorm(x) ~ u^2 - z^2 */
+    long s = signe(u);
+    if (s == -signe(v)) /* conjugate expression avoids cancellation */
+    {
+      z = gdiv(quadnorm(x), gsub(u, z));
+      if (b) shiftr_inplace(z, 1);
+    }
+    else
+    {
+      if (s) z = gadd(u, z);
+      if (b) shiftr_inplace(z, -1);
+    }
+  }
+  return gerepileupto(av, z);
 }
 
 static GEN
@@ -2415,9 +2484,9 @@ cvstop2(long s, GEN y)
 {
   GEN z, p = gel(y,2);
   long v, d = signe(gel(y,4))? precp(y): 0;
-  if (!s) return zeropadic(p, d);
+  if (!s) return zeropadic_shallow(p, d);
   v = z_pvalrem(s, p, &s);
-  if (d <= 0) return zeropadic(p, v);
+  if (d <= 0) return zeropadic_shallow(p, v);
   z = cgetg(5, t_PADIC);
   z[1] = evalprecp(d) | evalvalp(v);
   gel(z,2) = p;
@@ -2425,23 +2494,27 @@ cvstop2(long s, GEN y)
   gel(z,4) = modsi(s, gel(y,3)); return z;
 }
 
+static GEN
+itop2_coprime(GEN x, GEN y, long v, long d)
+{
+  GEN z = cgetg(5, t_PADIC);
+  z[1] = evalprecp(d) | evalvalp(v);
+  gel(z,2) = gel(y,2);
+  gel(z,3) = gel(y,3);
+  gel(z,4) = modii(x, gel(y,3)); return z;
+}
 /* cvtop(x, gel(y,2), precp(y)), shallow */
 GEN
 cvtop2(GEN x, GEN y)
 {
-  GEN z, p = gel(y,2);
+  GEN p = gel(y,2);
   long v, d = signe(gel(y,4))? precp(y): 0;
   switch(typ(x))
   {
     case t_INT:
-      if (!signe(x)) return zeropadic(p, d);
-      if (d <= 0) return zeropadic(p, Z_pval(x,p));
-      v = Z_pvalrem(x, p, &x);
-      z = cgetg(5, t_PADIC);
-      z[1] = evalprecp(d) | evalvalp(v);
-      gel(z,2) = p;
-      gel(z,3) = gel(y,3);
-      gel(z,4) = modii(x, gel(y,3)); return z;
+      if (!signe(x)) return zeropadic_shallow(p, d);
+      if (d <= 0) return zeropadic_shallow(p, Z_pval(x,p));
+      v = Z_pvalrem(x, p, &x); return itop2_coprime(x, y, v, d);
 
     case t_INTMOD:
       v = Z_pval(gel(x,1),p); if (v > d) v = d;
@@ -2450,18 +2523,18 @@ cvtop2(GEN x, GEN y)
     case t_FRAC:
     {
       GEN num, den;
-      if (d <= 0) return zeropadic(p, Q_pval(x,p));
+      if (d <= 0) return zeropadic_shallow(p, Q_pval(x,p));
       num = gel(x,1); v = Z_pvalrem(num, p, &num);
       den = gel(x,2); if (!v) v = -Z_pvalrem(den, p, &den);
-      z = cgetg(5, t_PADIC);
-      z[1] = evalprecp(d) | evalvalp(v);
-      gel(z,2) = p;
-      gel(z,3) = gel(y,3);
       if (!is_pm1(den)) num = mulii(num, Fp_inv(den, gel(y,3)));
-      gel(z,4) = modii(num, gel(y,3)); return z;
+      return itop2_coprime(num, y, v, d);
     }
     case t_COMPLEX: return ctop(x, p, d);
     case t_QUAD:    return qtop(x, p, d);
+    case t_PADIC:
+      if (!signe(gel(x,4))) return zeropadic_shallow(p, d);
+      if (precp(x) <= d) return x;
+      return itop2_coprime(gel(x,4), y, valp(x), d); /* reduce accuracy */
   }
   pari_err_TYPE("cvtop2",x);
   return NULL; /* LCOV_EXCL_LINE */
@@ -2533,9 +2606,7 @@ gcvtop(GEN x, GEN p, long r)
       for (i=2; i<lx; i++) gel(y,i) = gcvtop(gel(x,i),p,r);
       return y;
     case t_POLMOD: case t_RFRAC: case t_VEC: case t_COL: case t_MAT:
-      y = cgetg_copy(x, &lx);
-      for (i=1; i<lx; i++) gel(y,i) = gcvtop(gel(x,i),p,r);
-      return y;
+      pari_APPLY_same(gcvtop(gel(x,i),p,r));
   }
   return cvtop(x,p,r);
 }
@@ -2599,12 +2670,11 @@ sizedigit(GEN x)
 
 /* normalize series. avma is not updated */
 GEN
-normalize(GEN x)
+normalizeser(GEN x)
 {
   long i, lx = lg(x), vx=varn(x), vp=valp(x);
   GEN y, z;
 
-  if (typ(x) != t_SER) pari_err_TYPE("normalize",x);
   if (lx == 2) { setsigne(x,0); return x; }
   if (lx == 3) {
     z = gel(x,2);
