@@ -14,6 +14,9 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 #include "pari.h"
 #include "paripriv.h"
+
+#define DEBUGLEVEL DEBUGLEVEL_mathnf
+
 /**************************************************************/
 /**                                                          **/
 /**                HERMITE NORMAL FORM REDUCTION             **/
@@ -555,7 +558,7 @@ hnfspec(GEN mat, GEN perm, GEN* ptdep, GEN* ptB, GEN* ptC, long k0)
 {
   pari_sp av = avma;
   GEN H = hnfspec_i(mat, perm, ptdep, ptB, ptC, k0);
-  gerepileall(av, 4, ptC, ptdep, ptB, &H); return H;
+  return gc_all(av, 4, &H, ptC, ptdep, ptB);
 }
 
 /* HNF reduce x, apply same transforms to C */
@@ -660,7 +663,7 @@ hnfadd(GEN H, GEN perm, GEN* ptdep, GEN* ptB, GEN* ptC, /* cf hnfspec */
 {
   pari_sp av = avma;
   H = hnfadd_i(H, perm, ptdep, ptB, ptC, ZM_to_zm(extramat), extraC);
-  gerepileall(av, 4, ptC, ptdep, ptB, &H); return H;
+  return gc_all(av, 4, &H, ptC, ptdep, ptB);
 }
 
 /* zero aj = Aij (!= 0)  using  ak = Aik (maybe 0), via linear combination of
@@ -2053,8 +2056,7 @@ ZM_hnfall(GEN A, GEN *ptB, long remove)
 {
   pari_sp av = avma;
   A = ZM_hnfall_i(A, ptB, remove);
-  gerepileall(av, ptB? 2: 1, &A, ptB);
-  return A;
+  return gc_all(av, ptB? 2: 1, &A, ptB);
 }
 
 GEN
@@ -2495,6 +2497,13 @@ smithall(GEN x)
   return z;
 }
 
+GEN
+ZV_snfclean(GEN d)
+{
+  long c, l = lg(d);
+  for (c = 1; c < l; c++) { GEN t = gel(d,c); if (is_pm1(t)) break; }
+  return c == l? d: vec_shorten(d, c-1);
+}
 void
 ZM_snfclean(GEN d, GEN u, GEN v)
 {
@@ -2789,11 +2798,7 @@ snf_group(GEN H, GEN D, GEN *newU, GEN *newUi)
     }
     Ui = Hvec? ZM_diag_mul(H, V): ZM_mul(H, V);
     for (i = 1; i < l; i++) gel(Ui,i) = ZC_Z_divexact(gel(Ui,i), gel(D,i));
-    if (Hvec)
-    { for (i = 1; i < l; i++) gel(Ui,i) = vecmodii(gel(Ui,i), H); }
-    else
-      Ui = ZM_hnfrem(Ui, H);
-    *newUi = Ui;
+    *newUi = Hvec? ZM_ZV_mod(Ui, H): ZM_hnfrem(Ui, H);
   }
   return D;
 }
@@ -2888,3 +2893,81 @@ ZV_snf_trunc(GEN D)
   for (i = 1; i < l; i++)
     if (is_pm1(gel(D,i))) { setlg(D,i); break; }
 }
+
+long
+zv_snf_rank(GEN D, ulong p)
+{
+  long i, l = lg(D);
+  if (!p) return l - 1;
+  for (i = 1; i < l; i++)
+    if (D[i] % p) break;
+  return i - 1;
+}
+long
+ZV_snf_rank_u(GEN D, ulong p)
+{
+  long i, l = lg(D);
+  if (!p) return l - 1;
+  if (p == 2)
+  {
+    for (i = 1; i < l; i++)
+      if (mpodd(gel(D,i))) break;
+  }
+  else if (!(p & (p-1)))
+  { /* power of 2 */
+    long n = vals(p);
+    for (i = 1; i < l; i++)
+      if (umodi2n(gel(D,i), n)) break;
+  }
+  else
+  {
+    for (i = 1; i < l; i++)
+      if (umodiu(gel(D,i), p)) break;
+  }
+  return i - 1;
+}
+long
+ZV_snf_rank(GEN D, GEN p)
+{
+  long i, l;
+  switch(lgefint(p))
+  {
+    case 2: return lg(D) - 1;
+    case 3: return ZV_snf_rank_u(D, p[2]);
+  }
+  l = lg(D);
+  for (i = 1; i < l; i++)
+    if (!dvdii(gel(D,i), p)) break;
+  return i - 1;
+}
+long
+snfrank(GEN D, GEN p)
+{
+  long i, l;
+  if (typ(D) != t_VEC) pari_err_TYPE("snfrank", D);
+  l = lg(D);
+  if (l == 4 && typ(gel(D,3)) == t_MAT)
+  { /* from matsnf(,1) */
+    pari_sp av = avma;
+    long z;
+    GEN v;
+    D = gel(D,3); l = lg(D);
+    if (l == 1) return 0;
+    z = lgcols(D) - l; /* missing columns of 0s */
+    if (z < 0) pari_err_TYPE("snfrank", D);
+    v = cgetg(l, t_VEC);
+    for (i = 1; i < l; i++) gel(v, i) = gcoeff(D, i + z, i);
+    return gc_long(av, z + snfrank(v, p));
+  }
+  switch(typ(p))
+  {
+    case t_INT:
+      if (!RgV_is_ZV(D)) pari_err_TYPE("snfrank", D);
+      return ZV_snf_rank(D, p);
+    case t_POL: break;
+    default: pari_err_TYPE("snfrank", p);
+  }
+  for (i = 1; i < l; i++)
+    if (!gdvd(gel(D,i), p)) break;
+  return i - 1;
+}
