@@ -21,6 +21,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_polroots
+
 static const double pariINFINITY = 1./0.;
 
 static long
@@ -53,6 +55,7 @@ ZX_to_ZiX(GEN Pr, GEN Pi)
 {
   long i, lr = lg(Pr), li = lg(Pi), l = maxss(lr, li), m = minss(lr, li);
   GEN P = cgetg(l, t_POL);
+  P[1] = Pr[1];
   for(i = 2; i < m; i++)
     gel(P,i) = signe(gel(Pi,i)) ? mkcomplex(gel(Pr,i), gel(Pi,i))
                                 : gel(Pr,i);
@@ -60,7 +63,7 @@ ZX_to_ZiX(GEN Pr, GEN Pi)
     gel(P,i) = gel(Pr, i);
   for(     ; i < li; i++)
     gel(P,i) = mkcomplex(gen_0, gel(Pi, i));
-  return P;
+  return normalizepol_lg(P, l);
 }
 
 static GEN
@@ -220,21 +223,21 @@ static long
 newton_polygon(GEN p, long k)
 {
   pari_sp av = avma;
-  double *logcoef, slope;
-  long n = degpol(p), i, j, h, l, *vertex;
-
-  logcoef = (double*)stack_malloc_align((n+1)*sizeof(double), sizeof(double));
-  vertex = (long*)new_chunk(n+1);
+  long n = degpol(p), i, j, h, l, *vertex = (long*)new_chunk(n+1);
+  double *L = (double*)stack_malloc_align((n+1)*sizeof(double), sizeof(double));
 
   /* vertex[i] = 1 if i a vertex of convex hull, 0 otherwise */
-  for (i=0; i<=n; i++) { logcoef[i] = dbllog2(gel(p,2+i)); vertex[i] = 0; }
+  for (i=0; i<=n; i++) { L[i] = dbllog2(gel(p,2+i)); vertex[i] = 0; }
   vertex[0] = 1; /* sentinel */
   for (i=0; i < n; i=h)
   {
-    slope = logcoef[i+1]-logcoef[i];
-    for (j = h = i+1; j<=n; j++)
+    double slope;
+    h = i+1;
+    while (L[i] == -pariINFINITY) { i = h; h = i+1; vertex[h] = 1; }
+    slope = L[h] - L[i];
+    for (j = i+2; j<=n; j++) if (L[j] != -pariINFINITY)
     {
-      double pij = (logcoef[j]-logcoef[i])/(double)(j-i);
+      double pij = (L[j] - L[i])/(double)(j - i);
       if (slope < pij) { slope = pij; h = j; }
     }
     vertex[h] = 1;
@@ -242,7 +245,7 @@ newton_polygon(GEN p, long k)
   h = k;   while (!vertex[h]) h++;
   l = k-1; while (!vertex[l]) l--;
   set_avma(av);
-  return (long)floor((logcoef[h]-logcoef[l])/(double)(h-l) + 0.5);
+  return (long)floor((L[h]-L[l])/(double)(h-l) + 0.5);
 }
 
 /* change z into z*2^e, where z is real or complex of real */
@@ -505,7 +508,7 @@ logmin_modulus(GEN p, double tau)
 {
   pari_sp av = avma;
   if (gequal0(gel(p,2))) return -pariINFINITY;
-  return gc_double(av, - logmax_modulus(RgX_recip_shallow(p),tau));
+  return gc_double(av, - logmax_modulus(RgX_recip_i(p),tau));
 }
 
 /* return the log of the k-th modulus (ascending order) of p, rel. error tau*/
@@ -943,7 +946,7 @@ dft(GEN p, long k, long NN, long Lmax, long bit, GEN F, GEN H, long polreal)
   {
     aux=gel(U,k-i);
     for (j=1+i; j<k; j++) aux = gadd(aux,gmul(gel(F,2+j),gel(U,j-i)));
-    gel(H,i+2) = gdivgs(aux,NN);
+    gel(H,i+2) = gdivgu(aux,NN);
   }
 }
 
@@ -1064,9 +1067,9 @@ optimize_split(GEN p, long k, double delta, long bit,
     split_fromU(p,k,delta,bit,F,G,param,param2);
   else
   {
-    split_fromU(RgX_recip_shallow(p),n-k,delta,bit,&FF,&GG,param,param2);
-    *F = RgX_recip_shallow(GG);
-    *G = RgX_recip_shallow(FF);
+    split_fromU(RgX_recip_i(p),n-k,delta,bit,&FF,&GG,param,param2);
+    *F = RgX_recip_i(GG);
+    *G = RgX_recip_i(FF);
   }
 }
 
@@ -1128,7 +1131,7 @@ conformal_pol(GEN p, GEN a)
   Q = conformal_pol(Q, a);
   R = conformal_pol(R, a);
   S = gpowgs(deg1pol_shallow(gen_1, gneg(a), v), d);
-  T = RgX_recip_shallow(S);
+  T = RgX_recip_i(S);
   if (typ(a) == t_COMPLEX) T = gconj(T);
   if (odd(d)) T = RgX_neg(T);
   /* S = (X - a)^d, T = (conj(a) X - 1)^d */
@@ -1464,13 +1467,13 @@ split_0(GEN p, long bit, GEN *F, GEN *G)
     if (lr < LOG1_9) split_0_1(p, bit, F, G);
     else
     {
-      q = RgX_recip_shallow(p);
+      q = RgX_recip_i(p);
       lr = logmax_modulus(q,0.05);
       if (lr < LOG1_9)
       {
         split_0_1(q, bit, F, G);
-        *F = RgX_recip_shallow(*F);
-        *G = RgX_recip_shallow(*G);
+        *F = RgX_recip_i(*F);
+        *G = RgX_recip_i(*G);
       }
       else
         split_2(p,bit,NULL, 1.2837,F,G);
@@ -1728,69 +1731,60 @@ mygprec_special(GEN x, long bit)
 }
 
 static GEN
-fix_roots1(GEN r)
+fix_roots1(GEN R)
 {
-  long i, l = lg(r);
-  GEN allr = cgetg(l, t_VEC);
-  for (i=1; i<l; i++)
-  {
-    GEN t = gel(r,i);
-    gel(allr,i) = gcopy(t);
-    gunclone(t);
-  }
-  return allr;
+  long i, l = lg(R);
+  GEN v = cgetg(l, t_VEC);
+  for (i=1; i < l; i++) { GEN r = gel(R,i); gel(v,i) = gcopy(r); gunclone(r); }
+  return v;
 }
 static GEN
-fix_roots(GEN r, long h, long bit)
+fix_roots(GEN R, long h, long bit)
 {
-  long i, j, k, l, prec;
-  GEN allr, ro1;
-  if (h == 1) return fix_roots1(r);
-  prec = nbits2prec(bit);
-  ro1 = grootsof1(h, prec) + 1;
-  l = lg(r)-1;
-  allr = cgetg(h*l+1, t_VEC);
-  for (k=1,i=1; i<=l; i++)
+  long i, j, c, n, prec;
+  GEN v, Z, gh;
+
+  if (h == 1) return fix_roots1(R);
+  prec = nbits2prec(bit); Z = grootsof1(h, prec); gh = utoipos(h);
+  n = lg(R)-1; v = cgetg(h*n + 1, t_VEC);
+  for (c = i = 1; i <= n; i++)
   {
-    GEN p2, p1 = gel(r,i);
-    p2 = (h == 2)? gsqrt(p1, prec): gsqrtn(p1, utoipos(h), NULL, prec);
-    for (j=0; j<h; j++) gel(allr,k++) = gmul(p2, gel(ro1,j));
-    gunclone(p1);
+    GEN s, r = gel(R,i);
+    s = (h == 2)? gsqrt(r, prec): gsqrtn(r, gh, NULL, prec);
+    for (j = 1; j <= h; j++) gel(v, c++) = gmul(s, gel(Z,j));
+    gunclone(r);
   }
-  return allr;
+  return v;
 }
 
 static GEN
 all_roots(GEN p, long bit)
 {
-  GEN pd, q, roots_pol, m;
-  long bit2, i, e, h, elc, n = degpol(p);
-  double fb;
+  long bit2, i, e, h, n = degpol(p), elc = gexpo(leading_coeff(p));
+  GEN q, R, m, pd = RgX_deflate_max(p, &h);
+  double fb = fujiwara_bound(pd);
   pari_sp av;
 
-  pd = RgX_deflate_max(p, &h); elc = gexpo(leading_coeff(pd));
-  fb = fujiwara_bound(pd);
-  e = (fb < 0)? 0: (long)(2 * fb);
-  bit2 = bit + maxss(0, gexpo(pd)) + (long)log2(n/h)+1+e;
-  e = 0;
-  for (av=avma,i=1;; i++,set_avma(av))
+  if (fb < 0) fb = 0;
+  bit2 = bit + maxss(gexpo(p), 0) + (long)ceil(log2(n / h) + 2 * fb);
+  for (av = avma, i = 1, e = 0;; i++, set_avma(av))
   {
-    roots_pol = vectrunc_init(n+1);
+    R = vectrunc_init(n+1);
     bit2 += e + (n << i);
     q = RgX_gtofp_bit(mygprec(pd,bit2), bit2);
     q[1] = evalsigne(1)|evalvarn(0);
-    m = split_complete(q,bit2,roots_pol);
-    roots_pol = fix_roots(roots_pol, h, bit2);
+    m = split_complete(q, bit2, R);
+    R = fix_roots(R, h, bit2);
     q = mygprec_special(pd,bit2);
     q[1] = evalsigne(1)|evalvarn(0);
     e = gexpo(RgX_sub(q, m)) - elc + (long)log2((double)n) + 1;
-    if (e < -2*bit2) e = -2*bit2; /* avoid e = -oo */
     if (e < 0)
     {
-      e = bit + a_posteriori_errors(p,roots_pol,e);
-      if (e < 0) return roots_pol;
+      if (e < -2*bit2) e = -2*bit2; /* avoid e = -oo */
+      e = bit + a_posteriori_errors(p, R, e);
+      if (e < 0) return R;
     }
-    if (DEBUGLEVEL > 7)
+    if (DEBUGLEVEL)
       err_printf("all_roots: restarting, i = %ld, e = %ld\n", i,e);
   }
 }
@@ -1807,6 +1801,7 @@ isexactpol(GEN p)
   return 1;
 }
 
+/* p(0) != 0 [for efficiency] */
 static GEN
 solve_exact_pol(GEN p, long bit)
 {
@@ -2486,7 +2481,7 @@ ZX_Uspensky(GEN P, GEN ab, long flag, long bitprec)
     else
     { P = ZX_rescale(P, d); ad = gmul(a, d); bd = gmul(b, d); }
     diff = subii(bd, ad);
-    P = ZX_unscale(ZX_translate(P, ad), diff);
+    P = ZX_affine(P, diff, ad);
     res = usp(P, flag, bitprec);
     if (flag <= 1)
     {
