@@ -131,18 +131,33 @@ foreachpari(GEN x, GEN code)
   pop_lex(1); clone_unlock_deep(x);
 }
 
+/* is it better to sieve [a,b] or to factor individually ? */
+static int
+no_sieve(ulong a, ulong b)
+{ return b - a < usqrt(b) / tridiv_boundu(b); }
+
 /* 0 < a <= b. Using small consecutive chunks to 1) limit memory use, 2) allow
  * cheap early abort */
 static int
 forfactoredpos(ulong a, ulong b, GEN code)
 {
-  const ulong step = 1024;
+  ulong x1, step = maxuu(2 * usqrt(b), 1024);
   pari_sp av = avma;
-  ulong x1;
+  if (no_sieve(a, b))
+  {
+    ulong n;
+    for (n = a; n <= b; n++, set_avma(av))
+    {
+      GEN m = factoru(n);
+      set_lex(-1, mkvec2(utoipos(n), Flm_to_ZM(m)));
+      closure_evalvoid(code); if (loop_break()) return 1;
+    }
+    return 0;
+  }
   for(x1 = a;; x1 += step, set_avma(av))
   { /* beware overflow, fuse last two bins (avoid a tiny remainder) */
     ulong j, lv, x2 = (b >= 2*step && b - 2*step >= x1)? x1-1 + step: b;
-    GEN v = vecfactoru(x1, x2);
+    GEN v = vecfactoru_i(x1, x2);
     lv = lg(v);
     for (j = 1; j < lv; j++)
     {
@@ -175,9 +190,21 @@ zv_to_mZM(GEN v)
 static void
 forsquarefreepos(ulong a, ulong b, GEN code)
 {
-  const ulong step = 1024;
+  const ulong step = maxuu(1024, 2 * usqrt(b));
   pari_sp av = avma;
   ulong x1;
+  if (no_sieve(a, b))
+  {
+    ulong n;
+    for (n = a; n <= b; n++, set_avma(av))
+    {
+      GEN m = factoru(n);
+      if (!uissquarefree_fact(m)) continue;
+      set_lex(-1, mkvec2(utoipos(n), Flm_to_ZM(m)));
+      closure_evalvoid(code); if (loop_break()) return;
+    }
+    return;
+  }
   for(x1 = a;; x1 += step, set_avma(av))
   { /* beware overflow, fuse last two bins (avoid a tiny remainder) */
     ulong j, lv, x2 = (b >= 2*step && b - 2*step >= x1)? x1-1 + step: b;
@@ -197,9 +224,21 @@ forsquarefreepos(ulong a, ulong b, GEN code)
 static void
 forsquarefreeneg(ulong a, ulong b, GEN code)
 {
-  const ulong step = 1024;
+  const ulong step = maxuu(1024, 2 * usqrt(b));
   pari_sp av = avma;
   ulong x2;
+  if (no_sieve(a, b))
+  {
+    ulong n;
+    for (n = b; n >= a; n--, set_avma(av))
+    {
+      GEN m = factoru(n);
+      if (!uissquarefree_fact(m)) continue;
+      set_lex(-1, mkvec2(utoineg(n), zv_to_mZM(gel(m,1))));
+      closure_evalvoid(code); if (loop_break()) return;
+    }
+    return;
+  }
   for(x2 = b;; x2 -= step, set_avma(av))
   { /* beware overflow, fuse last two bins (avoid a tiny remainder) */
     ulong j, x1 = (x2 >= 2*step && x2-2*step >= a)? x2+1 - step: a;
@@ -257,19 +296,30 @@ Flm2negfact(GEN v, GEN M)
 static int
 forfactoredneg(ulong a, ulong b, GEN code)
 {
-  const ulong step = 1024;
+  ulong x2, step = maxuu(2 * usqrt(b), 1024);
   GEN P, E, M;
   pari_sp av;
-  ulong x2;
 
   P = cgetg(18, t_COL); gel(P,1) = gen_m1;
   E = cgetg(18, t_COL); gel(E,1) = gen_1;
   M = mkmat2(P,E);
   av = avma;
-  for(x2 = b;; x2 -= step, set_avma(av))
+  if (no_sieve(a, b))
+  {
+    ulong n;
+    for (n = b; n >= a; n--, set_avma(av))
+    {
+      GEN m = factoru(n);
+      Flm2negfact(m, M);
+      set_lex(-1, mkvec2(utoineg(n), M));
+      closure_evalvoid(code); if (loop_break()) return 1;
+    }
+    return 0;
+  }
+  for (x2 = b;; x2 -= step, set_avma(av))
   { /* beware overflow, fuse last two bins (avoid a tiny remainder) */
     ulong j, x1 = (x2 >= 2*step && x2-2*step >= a)? x2+1 - step: a;
-    GEN v = vecfactoru(x1, x2);
+    GEN v = vecfactoru_i(x1, x2);
     for (j = lg(v)-1; j; j--)
     { /* run backward: from factor(x1..x2) to factor(-x2..-x1) */
       ulong n = x1-1 + j;
@@ -1316,67 +1366,110 @@ sumpos0(GEN a, GEN code, long flag, long prec)
 GEN
 zbrent(void *E, GEN (*eval)(void *, GEN), GEN a, GEN b, long prec)
 {
-  long sig, iter, itmax;
+  long sig, iter, itmax, bit, bit0;
   pari_sp av = avma;
-  GEN c, d, e, tol, fa, fb, fc;
+  GEN c, d, e, fa, fb, fc;
 
+  if (typ(a) == t_INFINITY && typ(b) != t_INFINITY) swap(a,b);
+  if (typ(a) == t_INFINITY && typ(b) == t_INFINITY)
+  {
+    long s = gsigne(eval(E, real_0(prec))), r = 0;
+    if (gidentical(gel(a,1), gel(b,1)))
+      pari_err_DOMAIN("solve", "a and b", "=", a, mkvec2(a, b));
+    a = real_m1(prec); /* domain = R */
+    b = real_1(prec);
+    for(;;)
+    {
+      fa = eval(E, a);
+      fb = eval(E, b);
+      if (gsigne(fa) != s)
+      {
+        if (r) b[1] = evalsigne(-1) | _evalexpo(r-1); else b = real_0(prec);
+        break;
+      }
+      if (gsigne(fb) != s)
+      {
+        if (r) a[1] = evalsigne(1) | _evalexpo(r-1); else a = real_0(prec);
+        break;
+      }
+      r++; setexpo(a, r); setexpo(b, r);
+    }
+    c = b;
+    goto SOLVE;
+  }
+  if (typ(b) == t_INFINITY)
+  { /* a real, b == [+-]oo */
+    long s, r, minf = inf_get_sign(b) < 0;
+    GEN inc;
+    if (typ(a) != t_REAL || realprec(a) < prec) a = gtofp(a, prec);
+    fa = eval(E, a);
+    s = gsigne(fa);
+    inc = minf ? real_m1(prec) : real_1(prec);
+    r = gsigne(a) ? expo(a) : 0;
+    for(;;)
+    {
+      setexpo(inc, r);
+      b = addrr(a, inc); fb = eval(E, b);
+      if (gsigne(fb) != s) break;
+      a = b; fa = fb; r++;
+    }
+    if (minf) { c = a; swap(a, b); swap(fa, fb);} else c = b;
+    goto SOLVE;
+  }
   if (typ(a) != t_REAL || realprec(a) < prec) a = gtofp(a, prec);
   if (typ(b) != t_REAL || realprec(b) < prec) b = gtofp(b, prec);
   sig = cmprr(b, a);
   if (!sig) return gerepileupto(av, a);
-  if (sig < 0) {c = a; a = b; b = c;} else c = b;
+  if (sig < 0) { c = a; swap(a, b); } else c = b;
   fa = eval(E, a);
   fb = eval(E, b);
   if (gsigne(fa)*gsigne(fb) > 0)
     pari_err_DOMAIN("solve", "f(a)f(b)", ">", gen_0, mkvec2(fa, fb));
-  itmax = prec2nbits(prec) * 2 + 1;
-  tol = real2n(5-prec2nbits(prec), LOWDEFAULTPREC);
-  fc = fb;
-  e = d = NULL; /* gcc -Wall */
+SOLVE:
+  bit0 = -prec2nbits(prec); bit = 3+bit0; itmax = 1 - 2*bit0;
+  fc = fb; e = d = NULL;
   for (iter = 1; iter <= itmax; ++iter)
-  {
-    GEN xm, tol1;
-    if (gsigne(fb)*gsigne(fc) > 0)
-    {
-      c = a; fc = fa; e = d = subrr(b, a);
-    }
+  { /* b = current best guess, a and c auxiliary points */
+    long bit2;
+    GEN m;
+    if (gsigne(fb)*gsigne(fc) > 0) { c = a; fc = fa; e = d = subrr(b, a); }
     if (gcmp(gabs(fc, 0), gabs(fb, 0)) < 0)
-    {
-      a = b; b = c; c = a; fa = fb; fb = fc; fc = fa;
-    }
-    tol1 = abscmprr(tol, b) > 0? sqrr(tol): mulrr(tol, absr(b));
-    xm = shiftr(subrr(c, b), -1);
-    if (abscmprr(xm, tol1) <= 0 || gequal0(fb)) break; /* SUCCESS */
+    { a = b; b = c; c = a; fa = fb; fb = fc; fc = fa; }
+    bit2 = bit-1 + maxss(bit, expo(b));
+    m = subrr(c, b); shiftr_inplace(m, -1);
+    if (expo(m) <= expo(b)+bit0 || gequal0(fb)) break; /*SUCCESS*/
 
-    if (abscmprr(e, tol1) >= 0 && gcmp(gabs(fa, 0), gabs(fb, 0)) > 0)
-    { /* attempt interpolation */
+    if (expo(e) > bit2 && gexpo(fa) > gexpo(fb))
+    { /* interpolation, m != 0, |f(c)| >= |f(b)|, f(b)f(c) < 0 */
       GEN min1, min2, p, q, s = gdiv(fb, fa);
-      if (cmprr(a, c) == 0)
+      if (a == c || equalrr(a,c))
       {
-        p = gmul2n(gmul(xm, s), 1);
+        p = gmul2n(gmul(m, s), 1);
         q = gsubsg(1, s);
       }
       else
       {
         GEN r = gdiv(fb, fc);
         q = gdiv(fa, fc);
-        p = gmul2n(gmul(gsub(q, r), gmul(xm, q)), 1);
-        p = gmul(s, gsub(p, gmul(gsub(b, a), gsubgs(r, 1))));
+        p = gmul2n(gmul(gsub(q, r), gmul(m, q)), 1);
+        p = gmul(s, gsub(p, gmul(subrr(b, a), gsubgs(r, 1))));
         q = gmul(gmul(gsubgs(q, 1), gsubgs(r, 1)), gsubgs(s, 1));
       }
       if (gsigne(p) > 0) q = gneg_i(q); else p = gneg_i(p);
-      min1 = gsub(gmulsg(3, gmul(xm,q)), gabs(gmul(q, tol1), 0));
+      min1 = gsub(gmulsg(3, gmul(m,q)), gmul2n(gabs(q,0), bit2));
       min2 = gabs(gmul(e, q), 0);
       if (gcmp(gmul2n(p, 1), gmin_shallow(min1, min2)) < 0)
         { e = d; d = gdiv(p, q); } /* interpolation OK */
       else
-        { d = xm; e = d; } /* failed, use bisection */
+        e = d = m; /* failed, use bisection */
     }
-    else { d = xm; e = d; } /* bound decreasing too slowly, use bisection */
+    else e = d = m; /* bound decreasing too slowly, use bisection */
     a = b; fa = fb;
-    if (gcmp(gabs(d, 0), tol1) > 0) b = gadd(b, d);
-    else if (gsigne(xm) > 0)      b = addrr(b, tol1);
-    else                          b = subrr(b, tol1);
+    if (d == m) { b = addrr(c, b); shiftr_inplace(b,-1); }
+    else if (gexpo(d) > bit2) b = gadd(b, d);
+    else if (gsigne(m) > 0) b = addrr(b, real2n(bit2, LOWDEFAULTPREC));
+    else                    b = subrr(b, real2n(bit2, LOWDEFAULTPREC));
+    if (equalrr(a, b)) break;
     if (realprec(b) < prec) b = rtor(b, prec);
     fb = eval(E, b);
   }
@@ -1694,7 +1787,7 @@ derivfunk(void *E, GEN (*eval)(void *, GEN, long), GEN x, GEN ind0, long prec)
     break;
   case t_RFRAC:
     ind = chk_ind(ind0,&M); if (!ind) return cgetg(1,t_VEC);
-    x = rfrac_to_ser(x, precdl+2 + M * (1+rfrac_val_deriv(x)));
+    x = rfrac_to_ser_i(x, precdl+2 + M * (1+rfrac_val_deriv(x)));
     xp = derivser(x);
     break;
   case t_SER:
@@ -1734,7 +1827,7 @@ derivfun(void *E, GEN (*eval)(void *, GEN, long), GEN x, long prec)
     x = RgX_to_ser(x, precdl+2+ (1 + RgX_val(xp)));
     break;
   case t_RFRAC:
-    x = rfrac_to_ser(x, precdl+2+ (1 + rfrac_val_deriv(x)));
+    x = rfrac_to_ser_i(x, precdl+2+ (1 + rfrac_val_deriv(x)));
     /* fall through */
   case t_SER:
     xp = derivser(x);
